"""
Default values and metadata for EPREM simulation parameters.

DO NOT EDIT THIS FILE. It was created via the default-parameters.py script,
which is included in the eprempy/scripts directory.

Run `python default-parameters.py -h` for more information.
"""


BASETYPES_H = {
    'AU': {
        'value': 14959787070000.0,
        'info': 'One astronomical unit.',
        'unit': 'cm',
        'type': float,
    },
    'C': {
        'value': 29979245800.0,
        'info': 'The speed of light.',
        'unit': 'cm/s',
        'type': float,
    },
    'DAY': {
        'value': 0.0057755183314369945,
        'info': 'The conversion from EPREM time steps to Julian days.',
        'unit': 'day',
        'type': float,
    },
    'DEG2RAD': {
        'value': 0.017453292519943295,
        'info': 'The conversion from degrees to radians.',
        'unit': '1',
        'type': float,
    },
    'EV': {
        'value': 1.6022e-12,
        'info': 'The conversion from eVs to ergs.',
        'unit': 'erg/eV',
        'type': float,
    },
    'F': {
        'value': 0,
        'info': 'False',
        'unit': None,
        'type': int,
    },
    'FCONVERT': {
        'value': 1.2379901472361203e-12,
        'info': 'The conversion from distribution to flux.',
        'unit': '1',
        'type': float,
    },
    'GEV': {
        'value': 0.0016022,
        'info': 'The conversion from GeVs to ergs.',
        'unit': 'erg/GeV',
        'type': float,
    },
    'HALF_PI': {
        'value': 1.5707963267948966,
        'info': 'The value of π/2.',
        'unit': None,
        'type': float,
    },
    'LARGEFLOAT': {
        'value': 1e+33,
        'info': 'A very large float value.',
        'unit': None,
        'type': float,
    },
    'LARGEINT': {
        'value': 2147483647,
        'info': 'A very large integer value.',
        'unit': None,
        'type': int,
    },
    'MAX_STRING_SIZE': {
        'value': 240,
        'info': '',
        'unit': None,
        'type': int,
    },
    'MEV': {
        'value': 1.6022e-06,
        'info': 'The conversion from MeVs to ergs.',
        'unit': 'erg/MeV',
        'type': float,
    },
    'MHD_B_NORM': {
        'value': 0.038771870111656996,
        'info': 'The normalization for magnetic fields.',
        'unit': '1',
        'type': float,
    },
    'MHD_COUPLED': {
        'value': 1,
        'info': 'Couple to external MHD',
        'unit': None,
        'type': int,
    },
    'MHD_DEFAULT': {
        'value': 0,
        'info': 'Use the default MHD solver.',
        'unit': None,
        'type': int,
    },
    'MHD_DENSITY_NORM': {
        'value': 1.0,
        'info': 'The normalization factor for density.',
        'unit': '1',
        'type': float,
    },
    'MP': {
        'value': 1.6726e-24,
        'info': 'The proton mass.',
        'unit': 'g',
        'type': float,
    },
    'MZERO': {
        'value': 0.9382461065754594,
        'info': 'The proton rest-mass energy in GeV.',
        'unit': 'GeV',
        'type': float,
    },
    'OM': {
        'value': 185327.43617160583,
        'info': 'The normalization for ion gyrofrequency.',
        'unit': '1',
        'type': float,
    },
    'PI': {
        'value': 3.141592653589793,
        'info': 'The value of π.',
        'unit': None,
        'type': float,
    },
    'Q': {
        'value': 4.80320425e-10,
        'info': 'The proton charge.',
        'unit': 'statC',
        'type': float,
    },
    'RAD2DEG': {
        'value': 57.29577951308232,
        'info': 'The conversion from radians to degrees.',
        'unit': '1',
        'type': float,
    },
    'RSAU': {
        'value': 0.0046524726370988385,
        'info': 'The number of solar radii per au.',
        'unit': '1',
        'type': float,
    },
    'RSKM': {
        'value': 696000.0,
        'info': 'The number of solar radii per km.',
        'unit': '1',
        'type': float,
    },
    'RSUN': {
        'value': 69600000000.0,
        'info': 'The value of the solar radius.',
        'unit': 'cm',
        'type': float,
    },
    'SMALLFLOAT': {
        'value': 1e-33,
        'info': 'A very small float value.',
        'unit': None,
        'type': float,
    },
    'T': {
        'value': 1,
        'info': 'True',
        'unit': None,
        'type': int,
    },
    'TAU': {
        'value': 499.0047838361564,
        'info': 'The canonical EPREM time scale.',
        'unit': 's',
        'type': float,
    },
    'THRESH': {
        'value': 0.025,
        'info': 'The threshold for perpendicular diffusion.',
        'unit': '1',
        'type': float,
    },
    'TWO_PI': {
        'value': 6.283185307179586,
        'info': 'The value of 2π.',
        'unit': None,
        'type': float,
    },
    'VOLT': {
        'value': 0.0033333,
        'info': 'The conversion from volts to statvolts.',
        'unit': '1',
        'type': float,
    },
}
"""Values and metadata for constants defined in `baseTypes.h`.

Each entry contains the numerical value (`'value'`), a brief description
(`'info'`), the metric unit (`'unit'`), and the built-in type (`'type'`) of each
defined constant.

Notes
-----
* A `'unit'` of `None` implies a non-physical parameter (e.g., a boolean flag or
  system path), whereas a `'unit'` of `'1'` implies a unitless physical
  parameter.

See Also
--------
`~parameter.Defaults`
    A mapping from parameter to default value, with value converted to its
    equivalent built-in type.

`~parameter.Interface`
    An aliased mapping from parameter to user-provided value, if available, or
    default value.
"""


CONFIGURATION_C = {
    'abundance': {
        'type': list,
        'defaultVal': None,
        'maxVal': '1.0',
        'minVal': '0.0',
        'aliases': [],
        'unit': None,
    },
    'adiabaticChangeAlg': {
        'type': int,
        'defaultVal': '1',
        'minVal': '1',
        'maxVal': '3',
        'aliases': [],
        'unit': None,
    },
    'adiabaticFocusAlg': {
        'type': int,
        'defaultVal': '1',
        'minVal': '1',
        'maxVal': '3',
        'aliases': [],
        'unit': None,
    },
    'aziSunStart': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'rad',
    },
    'boundaryFunctAmplitude': {
        'type': float,
        'defaultVal': '1.0',
        'minVal': 'SMALLFLOAT',
        'maxVal': 'LARGEFLOAT',
        'aliases': ['J0'],
        'unit': '1 / (cm^2 * s * sr * (MeV/nuc))',
    },
    'boundaryFunctBeta': {
        'type': float,
        'defaultVal': '2.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': ['beta'],
        'unit': '1',
    },
    'boundaryFunctE0': {
        'type': float,
        'defaultVal': 'config.boundaryFunctEr',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
    },
    'boundaryFunctEcutoff': {
        'type': float,
        'defaultVal': '1.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': ['E0'],
        'unit': 'MeV',
    },
    'boundaryFunctEr': {
        'type': float,
        'defaultVal': '1.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': ['Er', 'reference energy'],
        'unit': 'MeV',
    },
    'boundaryFunctGamma': {
        'type': float,
        'defaultVal': '2.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': ['gamma'],
        'unit': '1',
    },
    'boundaryFunctR0': {
        'type': float,
        'defaultVal': '1.0',
        'minVal': 'config.rScale',
        'maxVal': 'LARGEFLOAT',
        'aliases': ['R0', 'r0', 'reference radius'],
        'unit': 'au',
    },
    'boundaryFunctXi': {
        'type': float,
        'defaultVal': '1.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': ['xi'],
        'unit': '1',
    },
    'boundaryFunctionInitDomain': {
        'type': bool,
        'defaultVal': '1',
        'aliases': [],
        'unit': None,
    },
    'charge': {
        'type': list,
        'defaultVal': '[1.0]',
        'maxVal': 'LARGEFLOAT',
        'minVal': '1.0',
        'aliases': [],
        'unit': 'e',
    },
    'checkSeedPopulation': {
        'type': bool,
        'defaultVal': '1',
        'aliases': [],
        'unit': None,
    },
    'dampCooling': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
    },
    'dsh_min': {
        'type': float,
        'defaultVal': '5.0e-5',
        'minVal': 'SMALLFLOAT',
        'maxVal': 'LARGEFLOAT',
    },
    'dumpFreq': {
        'type': int,
        'defaultVal': '1',
        'minVal': '0',
        'maxVal': '1000000',
        'aliases': [],
        'unit': None,
    },
    'eMax': {
        'type': float,
        'defaultVal': '1000.0',
        'minVal': 'config.eMin',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'MeV/nuc',
    },
    'eMin': {
        'type': float,
        'defaultVal': '1.0',
        'minVal': 'SMALLFLOAT',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'MeV/nuc',
    },
    'epCalcStartTime': {
        'type': float,
        'defaultVal': 'config.simStartTime',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'day',
    },
    'epremDomain': {
        'type': bool,
        'defaultVal': '0',
        'aliases': [],
        'unit': None,
    },
    'epremDomainOutputTime': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'day',
    },
    'flowMag': {
        'type': float,
        'defaultVal': '400.0e5',
        'minVal': 'SMALLFLOAT',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'cm/s',
    },
    'focusingLimit': {
        'type': float,
        'defaultVal': '1.0',
        'minVal': '0.0',
        'maxVal': '1.0',
    },
    'idealShock': {
        'type': bool,
        'defaultVal': '0',
        'aliases': [],
        'unit': None,
    },
    'idealShockFalloff': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': None,
    },
    'idealShockGradient': {
        'type': float,
        'defaultVal': 'config.idealShockSharpness / config.idealShockScaleLength',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
    },
    'idealShockInitTime': {
        'type': float,
        'defaultVal': 'config.simStartTime',
        'minVal': 'config.simStartTime',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'day',
    },
    'idealShockJump': {
        'type': float,
        'defaultVal': '4.0',
        'minVal': 'SMALLFLOAT',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': '1',
    },
    'idealShockPhi': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'TWO_PI',
        'aliases': [],
        'unit': 'rad',
    },
    'idealShockPhiWidth': {
        'type': float,
        'defaultVal': 'config.idealShockWidth',
        'minVal': '0.0',
        'maxVal': 'PI',
        'aliases': [],
        'unit': 'rad',
    },
    'idealShockScale': {
        'type': float,
        'defaultVal': 'config.idealShockSharpness / config.idealShockScaleLength',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'au',
    },
    'idealShockScaleLength': {
        'type': float,
        'defaultVal': '0.0046491',
        'minVal': 'SMALLFLOAT',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'au',
    },
    'idealShockSharpness': {
        'type': float,
        'defaultVal': '1.0',
        'minVal': 'SMALLFLOAT',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': '1',
    },
    'idealShockSpeed': {
        'type': float,
        'defaultVal': '1500e5',
        'minVal': 'SMALLFLOAT',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'cm/s',
    },
    'idealShockTheta': {
        'type': float,
        'defaultVal': 'HALF_PI',
        'minVal': '0.0',
        'maxVal': 'PI',
        'aliases': [],
        'unit': 'rad',
    },
    'idealShockThetaWidth': {
        'type': float,
        'defaultVal': 'config.idealShockWidth',
        'minVal': '0.0',
        'maxVal': 'PI',
        'aliases': [],
        'unit': 'rad',
    },
    'idealShockWidth': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'PI',
        'aliases': [],
        'unit': 'rad',
    },
    'idw_p': {
        'type': float,
        'defaultVal': '3.0',
        'minVal': 'SMALLFLOAT',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': '1',
    },
    'interpDistance': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
    },
    'interpWeight': {
        'type': float,
        'defaultVal': '3.0',
        'minVal': 'SMALLFLOAT',
        'maxVal': 'LARGEFLOAT',
    },
    'kperxkpar': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': ['kper_kpar', 'kper/kpar', 'kper / kpar'],
        'unit': '1',
    },
    'lamo': {
        'type': float,
        'defaultVal': '1.0',
        'minVal': 'SMALLFLOAT',
        'maxVal': 'LARGEFLOAT',
        'aliases': ['lam0', 'lambda0'],
        'unit': 'au',
    },
    'mass': {
        'type': list,
        'defaultVal': '[1.0]',
        'maxVal': 'LARGEFLOAT',
        'minVal': '1.0',
        'aliases': [],
        'unit': 'nucleon',
    },
    'mfpInverseB': {
        'type': bool,
        'defaultVal': '0',
    },
    'mfpPower': {
        'type': float,
        'defaultVal': 'config.mfpRadialPower',
        'minVal': '-1.0 * LARGEFLOAT',
        'maxVal': 'LARGEFLOAT',
    },
    'mfpRadialPower': {
        'type': float,
        'defaultVal': '2.0',
        'minVal': '-1.0 * LARGEFLOAT',
        'maxVal': 'LARGEFLOAT',
        'aliases': ['mfp_radial_power'],
        'unit': '1',
    },
    'mhdBAu': {
        'type': float,
        'defaultVal': '1.60e-5',
        'minVal': 'SMALLFLOAT',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'G',
    },
    'mhdBConvert': {
        'type': float,
        'defaultVal': '1.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': '1',
    },
    'mhdCouple': {
        'type': bool,
        'defaultVal': '0',
        'aliases': [],
        'unit': None,
    },
    'mhdCoupledTime': {
        'type': bool,
        'defaultVal': '1',
        'aliases': [],
        'unit': None,
    },
    'mhdDensityAu': {
        'type': float,
        'defaultVal': '8.30',
        'minVal': 'SMALLFLOAT',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'cm^-3',
    },
    'mhdDigits': {
        'type': int,
        'defaultVal': '3',
        'minVal': '0',
        'maxVal': '32767',
        'aliases': [],
        'unit': None,
    },
    'mhdDirectory': {
        'type': str,
        'defaultVal': '',
        'aliases': [],
        'unit': None,
    },
    'mhdInitFromOuterBoundary': {
        'type': int,
        'defaultVal': '2',
        'minVal': '0',
        'maxVal': '2',
        'aliases': [],
        'unit': None,
    },
    'mhdInitRadius': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'au',
    },
    'mhdInitTimeStep': {
        'type': float,
        'defaultVal': '0.000011574074074',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'day',
    },
    'mhdNumFiles': {
        'type': int,
        'defaultVal': '0',
        'minVal': '0',
        'maxVal': '32767',
        'aliases': [],
        'unit': None,
    },
    'mhdRadialMax': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'au',
    },
    'mhdRadialMin': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'au',
    },
    'mhdRhoConvert': {
        'type': float,
        'defaultVal': '1.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': '1',
    },
    'mhdRotateSolution': {
        'type': bool,
        'defaultVal': '1',
        'aliases': [],
        'unit': None,
    },
    'mhdStartTime': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'day',
    },
    'mhdSteadyState': {
        'type': bool,
        'defaultVal': '1',
        'aliases': [],
        'unit': None,
    },
    'mhdTimeConvert': {
        'type': float,
        'defaultVal': '1.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': '1',
    },
    'mhdVConvert': {
        'type': float,
        'defaultVal': '1.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': '1',
    },
    'mhdVmin': {
        'type': float,
        'defaultVal': '50.0e5',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'cm/s',
    },
    'numColumnsPerFace': {
        'type': int,
        'defaultVal': '2',
        'minVal': '1',
        'maxVal': 'LARGEINT',
        'aliases': [],
        'unit': None,
    },
    'numEnergySteps': {
        'type': int,
        'defaultVal': '20',
        'minVal': '2',
        'maxVal': 'LARGEINT',
        'aliases': [],
        'unit': None,
    },
    'numEpSteps': {
        'type': int,
        'defaultVal': '30',
        'minVal': '1',
        'maxVal': 'LARGEINT',
        'aliases': [],
        'unit': None,
    },
    'numMuSteps': {
        'type': int,
        'defaultVal': '11',
        'minVal': '2',
        'maxVal': 'LARGEINT',
        'aliases': [],
        'unit': None,
    },
    'numNodesPerStream': {
        'type': int,
        'defaultVal': 'N_PROCS',
        'minVal': 'N_PROCS',
        'maxVal': 'LARGEINT',
        'aliases': [],
        'unit': None,
    },
    'numObservers': {
        'type': int,
        'defaultVal': '0',
        'minVal': '0',
        'maxVal': '1000',
        'aliases': [],
        'unit': '1',
    },
    'numRowsPerFace': {
        'type': int,
        'defaultVal': '2',
        'minVal': '1',
        'maxVal': 'LARGEINT',
        'aliases': [],
        'unit': None,
    },
    'numSpecies': {
        'type': int,
        'defaultVal': '1',
        'minVal': '1',
        'maxVal': '100',
        'aliases': [],
        'unit': None,
    },
    'obsPhi': {
        'type': list,
        'defaultVal': '[0.0]',
        'maxVal': 'TWO_PI',
        'minVal': '0.0',
        'aliases': [],
        'unit': 'rad',
    },
    'obsR': {
        'type': list,
        'defaultVal': '[config.rScale]',
        'maxVal': 'LARGEFLOAT',
        'minVal': 'config.rScale',
        'aliases': [],
        'unit': 'au',
    },
    'obsTheta': {
        'type': list,
        'defaultVal': '[0.0]',
        'maxVal': 'PI',
        'minVal': '0.0',
        'aliases': [],
        'unit': 'rad',
    },
    'omegaSun': {
        'type': float,
        'defaultVal': '0.001429813',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'rad * cm / (au * s)',
    },
    'outputFloat': {
        'type': bool,
        'defaultVal': '0',
        'aliases': [],
        'unit': None,
    },
    'outputFlux': {
        'type': bool,
        'defaultVal': '0',
        'aliases': [],
        'unit': None,
    },
    'parallelFlow': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
    },
    'pointLegacyPrefix': {
        'type': bool,
        'defaultVal': '0',
    },
    'pointObserverOutput': {
        'type': bool,
        'defaultVal': '0',
        'aliases': [],
        'unit': None,
    },
    'pointObserverOutputTime': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'day',
    },
    'preEruptionDuration': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'day',
    },
    'rScale': {
        'type': float,
        'defaultVal': 'RSAU',
        'minVal': 'SMALLFLOAT',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'au',
    },
    'rigidityPower': {
        'type': float,
        'defaultVal': 'third',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': ['rigidity_power'],
        'unit': '1',
    },
    'seedFunctionTest': {
        'type': bool,
        'defaultVal': '0',
    },
    'simStartTime': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'day',
    },
    'simStopTime': {
        'type': float,
        'defaultVal': 'config.simStartTime + config.tDel',
        'minVal': 'config.simStartTime',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'day',
    },
    'streamFluxOutput': {
        'type': bool,
        'defaultVal': '0',
        'aliases': [],
        'unit': None,
    },
    'streamFluxOutputTime': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'day',
    },
    'streamLegacyPrefix': {
        'type': bool,
        'defaultVal': '0',
    },
    'streamSpawnLocAzi': {
        'type': list,
        'defaultVal': '[0.0]',
        'maxVal': 'TWO_PI',
        'minVal': '0.0',
        'aliases': [],
        'unit': 'rad',
    },
    'streamSpawnLocZen': {
        'type': list,
        'defaultVal': '[0.0]',
        'maxVal': 'PI',
        'minVal': '0.0',
        'aliases': [],
        'unit': 'rad',
    },
    'switchbackAlpha': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'PI',
    },
    'switchbackBeta': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'TWO_PI',
    },
    'switchbackPhi': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'TWO_PI',
    },
    'switchbackPhiWidth': {
        'type': float,
        'defaultVal': 'config.switchbackWidth',
        'minVal': '0.0',
        'maxVal': 'PI',
    },
    'switchbackR': {
        'type': float,
        'defaultVal': 'config.rScale',
        'minVal': 'config.rScale',
        'maxVal': 'LARGEFLOAT',
    },
    'switchbackTheta': {
        'type': float,
        'defaultVal': 'HALF_PI',
        'minVal': '0.0',
        'maxVal': 'PI',
    },
    'switchbackThetaWidth': {
        'type': float,
        'defaultVal': 'config.switchbackWidth',
        'minVal': '0.0',
        'maxVal': 'PI',
    },
    'switchbackWidth': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'PI',
    },
    'switchbackXc': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': '+LARGEFLOAT',
    },
    'tDel': {
        'type': float,
        'defaultVal': '0.01041666666667',
        'minVal': 'SMALLFLOAT',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'day',
    },
    'unifiedOutput': {
        'type': bool,
        'defaultVal': '1',
        'aliases': [],
        'unit': None,
    },
    'unifiedOutputTime': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'day',
    },
    'unstructuredDomain': {
        'type': bool,
        'defaultVal': '0',
        'aliases': [],
        'unit': None,
    },
    'unstructuredDomainOutputTime': {
        'type': float,
        'defaultVal': '0.0',
        'minVal': '0.0',
        'maxVal': 'LARGEFLOAT',
        'aliases': [],
        'unit': 'day',
    },
    'useAdiabaticChange': {
        'type': bool,
        'defaultVal': '1',
        'aliases': [],
        'unit': None,
    },
    'useAdiabaticFocus': {
        'type': bool,
        'defaultVal': '1',
        'aliases': [],
        'unit': None,
    },
    'useBoundaryFunction': {
        'type': bool,
        'defaultVal': '1',
    },
    'useDegrees': {
        'type': bool,
        'defaultVal': '0',
        'aliases': [],
        'unit': None,
    },
    'useDrift': {
        'type': bool,
        'defaultVal': '0',
        'aliases': [],
        'unit': None,
    },
    'useEPBoundary': {
        'type': bool,
        'defaultVal': '1',
    },
    'useManualStreamSpawnLoc': {
        'type': bool,
        'defaultVal': '0',
        'aliases': [],
        'unit': None,
    },
    'useMhdSteadyStateDt': {
        'type': bool,
        'defaultVal': '1',
        'aliases': [],
        'unit': None,
    },
    'useParallelDiffusion': {
        'type': bool,
        'defaultVal': '1',
        'aliases': [],
        'unit': None,
    },
    'useShellDiffusion': {
        'type': bool,
        'defaultVal': '0',
        'aliases': [],
        'unit': None,
    },
    'warningsFile': {
        'type': str,
        'defaultVal': 'warningsXXX.txt',
        'aliases': [],
        'unit': None,
    },
}
"""Metadata for parameters defined in `configuration.c`.

Each entry contains the default value (`'default'`), the smallest acceptable
value ('`minimum`'), the largest acceptable value ('`maximum`'), as well as the
equivalent built-in type and metric unit of those values (`'type'` and
`'unit'`). Some entries also contain declared aliases for the parameter
('`aliases`').

Notes
-----
* This dictionary stores all values as strings because some values are defined
  with respect to others.
* A `'unit'` of `None` implies a non-physical parameter (e.g., a boolean flag or
  system path), whereas a `'unit'` of `'1'` implies a unitless physical
  parameter.

See Also
--------
`~parameter.Defaults`
    A mapping from parameter to default value, with value converted to its
    equivalent built-in type.

`~parameter.Interface`
    An aliased mapping from parameter to user-provided value, if available, or
    default value.
"""





import numbers
import typing

import numpy
import numpy.typing


ValueType = typing.TypeVar('ValueType', int, float, numbers.Real)
"""Type variable representing real numeric values.

This type variable is constrained to be one of the following standard types

- `int`
- `float`
- `numbers.Real`
"""


# NOTE: These are a WIP attempt to define a generic type variable that
# represents either
# - a single real value
# - an array of arbitrary depth that contains real values
_T = typing.TypeVar('_T')
_NestedSequence = typing.Union[
    typing.Sequence[_T],
    typing.Sequence["_NestedSequence[_T]"],
]
DataType = typing.Union[
    ValueType,
    _NestedSequence[ValueType],
    numpy.typing.ArrayLike,
]





"""
Runtime-checkable protocols for general quantities.
"""

import abc
import typing


@typing.runtime_checkable
class Ordered(typing.Protocol):
    """Protocol for objects that support relative ordering.

    Classes that implement this protocol must define the following methods

    - `__lt__`
    - `__le__`
    - `__gt__`
    - `__ge__`

    The most appropriate return type will often be `bool` for each method, but
    exceptions exist (cf. `numpy.ndarray`).
    """

    __slots__ = ()

    @abc.abstractmethod
    def __lt__(self, other): ...

    @abc.abstractmethod
    def __le__(self, other): ...

    @abc.abstractmethod
    def __gt__(self, other): ...

    @abc.abstractmethod
    def __ge__(self, other): ...


@typing.runtime_checkable
class Comparable(Ordered, typing.Protocol):
    """Protocol for comparable objects.

    Classes that implement this protocol must implement the `~Ordered` protocol,
    and must define the following methods

    - `__eq__`
    - `__ne__`

    The most appropriate return type will often be `bool` for each method, but
    exceptions exist (cf. `numpy.ndarray`).
    """

    __slots__ = ()

    @abc.abstractmethod
    def __eq__(self, other): ...

    @abc.abstractmethod
    def __ne__(self, other): ...


@typing.runtime_checkable
class Additive(typing.Protocol):
    """Protocol for additive objects.

    Classes that implement this protocol must define the following methods

    - `__add__`
    - `__radd__`
    - `__sub__`
    - `__rsub__`

    Each method may return whatever type is appropriate for the class.
    """

    __slots__ = ()

    @abc.abstractmethod
    def __add__(self, other): ...

    @abc.abstractmethod
    def __radd__(self, other): ...

    @abc.abstractmethod
    def __sub__(self, other): ...

    @abc.abstractmethod
    def __rsub__(self, other): ...


@typing.runtime_checkable
class Multiplicative(typing.Protocol):
    """Protocol for multiplicative objects.

    Classes that implement this protocol must define the following methods

    - `__mul__`
    - `__rmul__`
    - `__truediv__`
    - `__rtruediv__`

    Each method may return whatever type is appropriate for the class.

    Notes
    -----
    - The floor- and modular-division operators do not appear in this protocol,
      because their distinction from "true" division really only makes sense in
      the context of real-valued numeric objects, whereas a more general set of
      objects may implement this protocol (e.g., symbolic expressions).
    """

    __slots__ = ()

    @abc.abstractmethod
    def __mul__(self, other): ...

    @abc.abstractmethod
    def __rmul__(self, other): ...

    @abc.abstractmethod
    def __truediv__(self, other): ...

    @abc.abstractmethod
    def __rtruediv__(self, other): ...


@typing.runtime_checkable
class Algebraic(Additive, Multiplicative, typing.Protocol):
    """Protocol for algebraic objects.

    Classes that implement this protocol must implement the `~Additive` and
    `~Multiplicative` protocols, and must define the `__pow__` method, which may
    return whatever type is appropriate for the class.

    Notes
    -----
    - The formal definition of an algebraic quantity requires that
      exponentiation involve only constant, rational exponents. However, this
      protocol does not place any restrictions on the type of exponent.
    - This protocol differs from the `~Additive` and `~Multiplicative`
      protocols, which require foward and reverse versions of their operators,
      by not requiring `__rpow__`. Doing so would imply that any algebraic
      object should be allowed as an exponent, which is not generally true. Of
      course, this doesn't prevent a concrete object from defining `__rpow__`.
    """

    __slots__ = ()

    @abc.abstractmethod
    def __pow__(self, other): ...


@typing.runtime_checkable
class Complex(Algebraic, typing.Protocol):
    """Protocol for complex-valued objects.

    Classes that implement this protocol must implement the `~Algebraic`
    protocol, and must define the following methods

    - `__abs__`
    - `__pos__`
    - `__neg__`

    Each method may return whatever type is appropriate for the class.
    """

    __slots__ = ()

    @abc.abstractmethod
    def __abs__(self): ...

    @abc.abstractmethod
    def __pos__(self): ...

    @abc.abstractmethod
    def __neg__(self): ...


@typing.runtime_checkable
class Real(Comparable, Complex, typing.Protocol):
    """Protocol for real-valued numeric objects.

    Classes that implement this protocol must implement the `~Comparable` and
    `~Complex` protocols, and must define the following methods

    - `__rpow__`
    - `__floordiv__`
    - `__rfloordiv__`
    - `__mod__`
    - `__rmod__`

    Each method may return whatever type is appropriate for the class.
    """

    @abc.abstractmethod
    def __rpow__(self, other): ...

    @abc.abstractmethod
    def __floordiv__(self, other): ...

    @abc.abstractmethod
    def __rfloordiv__(self, other): ...

    @abc.abstractmethod
    def __mod__(self, other): ...

    @abc.abstractmethod
    def __rmod__(self, other): ...


@typing.runtime_checkable
class Scalar(Real, typing.Protocol):
    """Protocol for real-valued scalar objects.

    Classes that implement this protocol must implement the `~Real` protocol,
    and must define the following methods

    - `__int__`, which must return an `int`
    - `__float__`, which must return a `float`
    - `__round__`, which may return whatever type is appropriate to the class

    Notes
    -----
    - This protocol does not require `__trunc__` (called by `math.trunc`),
      `__floor__` (called by `math.floor`), or `__ceil__` (called by
      `math.ceil`) because doing so would exclude the built-in `float` type.
    """

    __slots__ = ()

    @abc.abstractmethod
    def __int__(self) -> int: ...

    @abc.abstractmethod
    def __float__(self) -> float: ...

    @abc.abstractmethod
    def __round__(self): ...


@typing.runtime_checkable
class Sequence(typing.Protocol):
    """Protocol for sequence-like objects.

    Classes that implement this protocol must define the following methods

    - `__contains__`, which should return a `bool`
    - `__len__`, which should return an `int`
    - `__iter__`, which should return an iterator over the value type
    - `__getitem__`, which may return an appropriate value type
    - `__array__`, which should return a `numpy.ndarray`
    """

    __slots__ = ()

    @abc.abstractmethod
    def __contains__(self, v, /): ...

    @abc.abstractmethod
    def __len__(self): ...

    @abc.abstractmethod
    def __iter__(self): ...

    @abc.abstractmethod
    def __getitem__(self, i, /): ...

    @abc.abstractmethod
    def __array__(self, *args, **kwargs): ...


@typing.runtime_checkable
class Array(Sequence, typing.Protocol):
    """Protocol for array-like objects.

    Classes that implement this protocol must implement the `~Sequence`
    protocol, and must define the following properties

    - `size`, which should return an `int`
    - `ndim`, which should return an `int`
    - `shape`, which should return a `tuple` of `int`s
    """

    __slots__ = ()

    @property
    @abc.abstractmethod
    def size(self) -> int: ...

    @property
    @abc.abstractmethod
    def ndim(self) -> int: ...

    @property
    @abc.abstractmethod
    def shape(self) -> typing.Tuple[int, ...]: ...


@typing.runtime_checkable
class HasData(typing.Protocol):
    """Protocol for objects with a `data` property.

    Classes that implement this protocol must define the `data` property.
    """

    __slots__ = ()

    @property
    @abc.abstractmethod
    def data(self): ...


@typing.runtime_checkable
class HasUnit(typing.Protocol):
    """Protocol for objects with a `unit` property.

    Classes that implement this protocol must define the `unit` property.
    """

    __slots__ = ()

    @property
    @abc.abstractmethod
    def unit(self): ...


@typing.runtime_checkable
class Variable(Sequence, typing.Protocol):
    """Protocol for variable objects.

    Classes that implement this protocol must implement the `~Sequence`
    protocol, must define the following methods

    - `__int__`, which must return an `int`
    - `__float__`, which must return a `float`
    - `__complex__`, which must return a `complex`

    and must define the `size` property, which should return an `int`.
    """

    __slots__ = ()

    @abc.abstractmethod
    def __int__(self) -> int: ...

    @abc.abstractmethod
    def __float__(self) -> float: ...

    @abc.abstractmethod
    def __complex__(self) -> complex: ...

    @property
    @abc.abstractmethod
    def size(self) -> int: ...


@typing.runtime_checkable
class Measured(HasData, HasUnit, typing.Protocol):
    """Protocol for measured objects.
    
    Classes that implement this protocol must implement the `~HasData` and
    `~HasUnit` protocols.
    """

    __slots__ = ()


@typing.runtime_checkable
class Measurement(Measured, Variable, typing.Protocol):
    """Protocol for measurement-like objects.
    
    Classes that implement this protocol must implement the `~Measured` and
    `~Variable` protocols.
    """

    __slots__ = ()


@typing.runtime_checkable
class Measurable(typing.Protocol):
    """Protocol for explicitly measurable objects.

    Classes that implement this protocol must define the `__measure__` method,
    which should return an object that implements the `~Measurement` protocol.

    See Also
    --------
    `~physical.measure`
        The function that calls `__measure__`.
    """

    __slots__ = ()

    @abc.abstractmethod
    def __measure__(self, **kwargs) -> Measurement: ...




import collections.abc
import functools
import itertools
import typing

import numpy
import numbers

from . import etc


T = typing.TypeVar('T')


@typing.overload
def unique(
    arg: typing.List[T],
    *,
    strict: bool=False,
) -> typing.List[T]: ...

@typing.overload
def unique(
    arg: T,
    *,
    strict: typing.Literal[True],
) -> typing.List[T]: ...

@typing.overload
def unique(*args: T, strict: bool=False) -> typing.List[T]: ...

def unique(*args, strict=False):
    """Remove repeated items from `args` while preserving order.
    
    Parameters
    ----------
    *args
        The items to compare.

    strict : bool, default=false
        By default, if `args` comprises a single iterable object, this function
        will extract that object under the assumption that the caller wants to
        remove repeated items from the given iterable object. If `strict` is
        true, this function will operate on the input as given.
    """
    items = (
        args[0] if (not strict and len(args) == 1)
        else args
    )
    collection = []
    for item in items:
        if item not in collection:
            collection.append(item)
    return collection


_Wrapped = typing.TypeVar('_Wrapped', bound=typing.Iterable)


@typing.overload
def unwrap(obj: typing.Union[T, typing.Iterable[T]]) -> T: ...


@typing.overload
def unwrap(
    obj: typing.Union[T, typing.Iterable[T]],
    newtype: typing.Type[_Wrapped]=None,
) -> _Wrapped: ...


def unwrap(obj, newtype=None):
    """Remove redundant outer lists and tuples.

    This function will strip away enclosing instances of `list` or `tuple`, as
    long as they contain a single item, until it finds an object of a different
    type, a `list` or `tuple` containing multiple items, or an empty `list` or
    `tuple`.

    Parameters
    ----------
    obj : Any
        The object to "unwrap".

    newtype : type
        An iterable type into which to store the result. Specifying this allows
        the caller to ensure that the result is an iterable object after
        unwrapping interior iterables.

    Returns
    -------
    Any
        The element enclosed in multiple instances of `list` or `tuple`, or a
        (possibly empty) `list` or `tuple`.

    Examples
    --------
    Unwrap numbers enclosed in increasingly deeper lists:

    >>> cases = [[1], [[2]], [[[3]]], [[[[4]]]]]
    >>> for case in cases:
    ...     print(unwrap(case))
    ... 
    1
    2
    3
    4

    It preserves numbers and strings that are already unwrapped:

    >>> unwrap(42)
    42
    >>> unwrap('string')
    'string'

    Passing a type to `newtype` ensures a result of that type:

    >>> unwrap(42, newtype=tuple)
    (42,)
    >>> unwrap(42, newtype=list)
    [42]
    >>> unwrap([42], newtype=list)
    [42]
    >>> unwrap(([(42,)],), newtype=list)
    [42]

    It works with multiple wrapped elements:

    >>> unwrap([1, 2])
    [1, 2]
    >>> unwrap([[1, 2]])
    [1, 2]
    >>> unwrap(['one', 'two'])
    ['one', 'two']
    >>> unwrap([['one', 'two']])
    ['one', 'two']

    It stops at an empty `list` or `tuple`:

    >>> unwrap([])
    []
    >>> unwrap(())
    ()
    >>> unwrap(list())
    []
    >>> unwrap(tuple())
    ()
    >>> unwrap([[]])
    []
    >>> unwrap([()])
    ()
    >>> unwrap([], newtype=tuple)
    ()
    """
    seed = [obj]
    wrapped = (list, tuple)
    while isinstance(seed, wrapped) and len(seed) == 1:
        seed = seed[0]
    if newtype is not None:
        return newtype(wrap(seed))
    return seed


def wrap(
    arg: typing.Optional[typing.Union[T, typing.Iterable[T]]],
) -> typing.List[T]:
    """Wrap `arg` in a list, if necessary.

    In most cases, this function will try to iterate over `arg`. If that
    operation succeeds, it will simply return `arg`; if the attempt to iterate
    raises a `TypeError`, it will assume that `arg` is a scalar and will return
    a one-element list containing `arg`. If `arg` is `None`, this function will
    return an empty list. If `arg` is a string, this function will return a
    one-element list containing `arg`.
    """
    if arg is None:
        return []
    if isinstance(arg, str):
        return [arg]
    try:
        iter(arg)
    except TypeError:
        return [arg]
    else:
        return list(arg)


class Wrapper(collections.abc.Collection, typing.Generic[T]):
    """A collection of independent members.

    This class represents an iterable collection with members that have meaning
    independent of any other members. When initialized with a "separable" object
    (e.g., a `list`, `tuple`, or `set`), the new instance will behave like the
    equivalent `tuple`. When initialized with a non-"separable" object, the new
    instance will behave like a `tuple` containing that object.

    See Also
    --------
    `~isseparable`
    """

    def __init__(
        self,
        this: typing.Optional[typing.Union[T, typing.Iterable[T]]],
        /,
    ) -> None:
        """Initialize a wrapped object from `this`"""
        self._arg = this
        self._wrapped = tuple(wrap(this))

    def __iter__(self) -> typing.Iterator[T]:
        """Called for iter(self)."""
        return iter(self._wrapped)

    def __len__(self) -> int:
        """Called for len(self)."""
        return len(self._wrapped)

    def __contains__(self, __x: object) -> bool:
        """Called for __x in self."""
        return __x in self._wrapped

    def __eq__(self, other) -> bool:
        """True if two wrapped objects have equal arguments."""
        if isinstance(other, Wrapper):
            return sorted(self) == sorted(other)
        return NotImplemented

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return self._arg


def isseparable(x, /):
    """True if `x` is iterable but is not string-like.

    This function identifies iterable collections with members that have meaning
    independent of any other members. For example, a list of numbers is
    "separable" whereas a string is not, despite the fact that both objects are
    iterable collections.

    The motivation for this distinction is to make it easier to treat single
    numbers and strings equivalently to iterables of numbers and strings.
    """
    try:
        iter(x)
    except TypeError:
        return False
    return not isinstance(x, (str, bytes))


def hastype(
    __obj,
    __types: typing.Union[type, typing.Tuple[type, ...]],
    *wrappers: typing.Type[typing.Iterable],
    strict: bool=False,
) -> bool:
    """True if an object is a certain type or contains certain types.
    
    Parameters
    ----------
    __obj : Any
        The object to compare.

    __types : type or tuple of types
        One or more types of which the target object may be an instance.

    *wrappers : iterable type
        Zero or more iterable types of which the target object may be an
        instance. If the target object is an instance of a given wrapper type,
        this function will test whether every member of the target object is an
        instance of the given types.

    strict : bool, default=False
        If true, return `True` if `__obj` contains only members with the target
        type(s). Otherwise, return `True` if `__obj` contains at least one
        member with the target type(s).

    Examples
    --------
    When called without wrappers, this function is identical to `isinstance`:

    >>> iterables.hastype(1, int)
    True

    >>> iterables.hastype('s', str)
    True

    >>> iterables.hastype([1, 2], list)
    True

    Note that in these cases, `strict` is irrelevant because this function
    checks only the type of `__obj`.

    The target object contains the given type but `list` is not a declared
    wrapper:
    
    >>> iterables.hastype([1, 2], int)
    False
    
    Same as above, but this time `list` is a known wrapper:

    >>> iterables.hastype([1, 2], int, list)
    True
    
    Similar, except only `tuple` is declared as a wrapper:

    >>> iterables.hastype([1, 2], int, tuple)
    False

    By default, only one member of a wrapped object needs to be an instance of
    one of the target types:

    >>> iterables.hastype([1, 2.0], int, list)
    True

    If `strict=True`, each member must be one of the target types:

    >>> iterables.hastype([1, 2.0], int, list, strict=True)
    False

    Multiple target types must be passed as a `tuple`, just as when calling
    `isinstance`:

    >>> iterables.hastype([1, 2.0], (int, float), list)
    True

    Otherwise, this function will interpret them as wrapper types:

    >>> iterables.hastype([1, 2.0], int, float, list, strict=True)
    False
    """
    if isinstance(__obj, __types):
        return True
    for wrapper in wrappers:
        if isinstance(__obj, wrapper):
            check = all if strict else any
            return check(isinstance(i, __types) for i in __obj)
    return False


def find_first(__type: typing.Type[T], *args):
    """Find the first argument with the given type."""
    for arg in args:
        if isinstance(arg, __type):
            return arg


@etc.str2repr
class Graph(collections.abc.Collection):
    """A collection of weighted edges."""

    def __init__(self, base: typing.Mapping=None) -> None:
        """Initialize an instance.

        Parameters
        ----------
        base : mapping, optional
            The mapping of weighted connections from which to initialize this
            instance. Items in the mapping must have the form `((start, end),
            weight)`.
        """
        self.connections: typing.Dict[typing.Tuple[str, str], float] = {}
        """The forward and reverse links in this graph."""
        base = base or {}
        for (start, end), weight in base.items():
            self.add_connection(start, end, weight)

    def __contains__(self, connection: typing.Tuple[str, str]):
        """True if `connection` is available."""
        return connection in self.connections

    def __len__(self) -> int:
        """The number of connections. Called for len(self)."""
        return len(self.connections)

    def __iter__(self):
        """Iterate over connections. Called for iter(self)."""
        return iter(self.connections)

    @property
    def nodes(self):
        """The distinct nodes in this graph."""
        return {n for connection in self.connections for n in connection}

    def get_adjacencies(self, node: str):
        """Retrieve the connections to this node.
        
        Parameters
        ----------
        node : string
            The key corresponding to the target node.

        Returns
        -------
        `~dict`
            A dictionary whose keys represent the nodes connected to `node` and
            whose values represent the corresponding edge weight. An empty
            dictionary represents a node with no connections.
        """
        return {
            end: v for (start, end), v in self.connections.items()
            if start == node
        } if node in self.nodes else {}

    def get_weight(self, start: str, end: str):
        """Retrieve the weight of this link, if possible."""
        if (start, end) in self.connections:
            return self.connections[(start, end)]
        raise KeyError(
            f"No connection between {start!r} and {end!r}"
        ) from None

    def add_connection(
        self,
        start: str,
        end: str,
        weight: float=None,
    ) -> None:
        """Add a connection (with optional weight) to the graph."""
        forward = ((start, end), weight)
        inverse = 1 / weight if weight else weight
        reverse = ((end, start), inverse)
        for edge, value in (forward, reverse):
            if edge not in self.connections:
                self.connections[edge] = value

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return '\n'.join(
            f"({n0} -> {n1}): {wt}"
            for (n0, n1), wt in self.connections.items()
        )


class TableKeyError(Exception):
    """No common key with this name."""
    def __str__(self) -> str:
        if len(self.args) > 0:
            return f"Table has no common key '{self.args[0]}'"
        return "Key not found in table"


class TableValueError(Exception):
    """An exception occurred during value-based look-up."""

    def __init__(self, value: typing.Any) -> None:
        self.value = value


class AmbiguousValueError(TableValueError):
    """Failed to find a unique entry by value."""

    def __str__(self) -> str:
        return f"No unique entry containing {self.value!r}"


class MissingValueError(TableValueError):
    """Failed to find any qualifying entries by value."""

    def __str__(self) -> str:
        return f"No entries containing {self.value!r}"


class TableRequestError(Exception):
    """An exception occurred during standard look-up."""

    def __init__(self, request: typing.Mapping) -> None:
        self.request = request


class AmbiguousRequestError(TableRequestError):
    """There are multiple instances of the same value for this key."""

    def __str__(self) -> str:
        """Print the requested pairs in the order provided."""
        items = list(self.request.items())
        n_items = len(items)
        if n_items <= 2:
            requested = " and ".join(f"'{k}={v}'" for k, v in items)
        else:
            these = ", ".join(f"'{k}={v}'" for k, v in items[:-1])
            this = f"'{items[-1][0]}={items[-1][1]}'"
            requested = f"{these}, and {this}"
        if n_items == 1:
            return f"The search criterion {requested} is ambiguous"
        return f"The search criteria {requested} are ambiguous"


class TableLookupError(TableRequestError):
    """Could not find the requested key-value pair(s)."""

    def __str__(self) -> str:
        """Print the requested pairs in the order provided."""
        items = list(self.request.items())
        if len(items) <= 2:
            criteria = " and ".join(f"{k}={v}" for k, v in items)
        else:
            these = ", ".join(f"{k}={v}" for k, v in items[:-1])
            this = f"{items[-1][0]}={items[-1][1]}"
            criteria = f"{these}, and {this}"
        return f"Table has no entry with {criteria}"


class Table(collections.abc.Mapping):
    """A collection of mappings with support for multi-key search.

    Subclasses of this class may override the `_parse` method to customize
    parsing user input into a key-value pair, and the `_prepare` method to
    modify the search result (e.g., cast it to a custom type) before returning.
    """

    _KT = typing.TypeVar('_KT', bound=str)

    _VT = typing.TypeVar('_VT')

    _ET = typing.TypeVar('_ET', bound=typing.Mapping)

    def __init__(self, entries: typing.Collection[_ET]) -> None:
        self._entries = entries
        self._keys = None

    def __len__(self) -> int:
        """Called for len(self)."""
        return len(self._entries)

    def __iter__(self) -> typing.Iterator[_ET]:
        """Called for iter(self)."""
        return iter(self._entries)

    def show(self, names: typing.Iterable[str]=None):
        """Print a formatted version of this table."""
        colpad = 2
        names = names or ()
        columns = {
            name: {'width': len(name)}
            for name in names if name in self.keys
        }
        dashed = {
            name: '-' * column['width']
            for name, column in columns.items()
        }
        for name, column in columns.items():
            values = self.get(name, ())
            for value in values:
                column['width'] = max(column['width'], len(str(value)))
        headers = [
            f"{name:^{2*colpad + column['width']}}"
            for name, column in columns.items()
        ]
        lines = [
            f"{dashed[name]:^{2*colpad + column['width']}}"
            for name, column in columns.items()
        ]
        rows = [
            [
                f"{entry[name]:^{2*colpad + column['width']}}"
                for name, column in columns.items()
            ] for entry in self
        ]
        if headers:
            print(''.join(header for header in headers if header))
        if lines:
            print(''.join(line for line in lines if line))
        for row in rows:
            if row:
                print(''.join(row))

    @property
    def keys(self) -> typing.Set[_KT]:
        """All the keys common to the individual mappings."""
        if self._keys is None:
            all_keys = [list(entry.keys()) for entry in self._entries]
            self._keys = set(all_keys[0]).intersection(*all_keys[1:])
        return self._keys

    def __getitem__(self, key: _KT) -> typing.Tuple[_VT]:
        """Get all the values for a given key if it is common."""
        if key in self.keys:
            values = [entry[key] for entry in self._entries]
            return tuple(values)
        raise TableKeyError(key)

    def get(self, key: _KT, default: typing.Any=None) -> typing.Tuple[_VT]:
        """Get all the values for a given key when available."""
        try:
            return self[key]
        except TableKeyError:
            values = [entry.get(key, default) for entry in self._entries]
            return tuple(values)

    def find(
        self,
        value: _VT,
        unique: bool=False,
    ) -> typing.Union[_ET, typing.List[_ET]]:
        """Find entries with the given value."""
        found = [entry for entry in self._entries if value in entry.values()]
        if not found:
            raise MissingValueError(value)
        if not unique:
            return found
        if len(found) > 1:
            raise AmbiguousValueError(value)
        return found[0]

    def __call__(self, strict: bool=False, **request):
        """Look up an entry by user-requested keys.

        This method will try to return the unique entry in the instance
        collection that contains the given key(s) with the corresponding
        value(s). The iterative search (`strict=False`; default) will iterate
        through the key-value pairs until it either finds a unique entry or runs
        out of pairs. The strict search (`strict=True`) will attempt to find the
        unique entry with all the key-value pairs. Both searches will raise a
        `TableLookupError` if they fail to find a unique entry, or a
        `TableKeyError` if one of the given keys does not exist in all table
        entries. The iterative search will raise an `AmbiguousRequestError` if
        the given key-value pairs are insufficient to determine a unique entry.

        Parameters
        ----------
        strict : bool, default=False
            Fail if any of the given key-value pairs is not in the collection.

        **request : mapping
            Key-value pairs that define the search criteria. Each key must
            appear in all table entries.

            Key-value pairs in which each requested key is a key in every entry
            of the collection. Each requested value may correspond to its
            respective key in zero or more entries for an iterative search, but
            must correspond to its respective key in at least one entry for a
            strict search.

        Returns
        -------
        mapping
            The unique entry containing at least one of the requested key-value
            pairs (iterative search) or all of the requested key-value pairs
            (strict search).

        Raises
        ------
        TableKeyError
            A requested key is not present in every entry.
        
        TableLookupError
            Could not find an entry that matched the requested key-value pairs.

        AmbiguousRequestError
            The given key-value pair is ambiguous. Only applies to iterative
            search.
        """
        subset = [*self._entries]
        for n_checked, pair in enumerate(request.items(), start=1):
            key, value = pair
            if key not in self.keys:
                raise TableKeyError(key)
            subset = [
                entry for entry in subset
                if entry[key] == value
            ]
            if not strict:
                count = self[key].count(value)
                if count > n_checked and len(request) == n_checked:
                    raise AmbiguousRequestError(request)
                if len(subset) == 1:
                    return subset[0]
        if strict and len(subset) == 1:
            return subset[0]
        raise TableLookupError(request)


def distribute(a, b):
    """Distribute `a` and `b` over each other.

    If both `a` and `b` are separable, this function will return their Cartesian
    product. If only `a` or `b` is separable, this function will pair the
    non-separable argument with each element of the separable argument. If
    neither is separable, this function will raise an error.

    See Also
    --------
    `~isseparable`
        Determine if an object is iterable but is not string-like.
    """
    a_separable = isseparable(a)
    b_separable = isseparable(b)
    if not (a_separable or b_separable):
        raise TypeError("At least one argument must be whole")
    if a_separable and b_separable:
        return iter(itertools.product(a, b))
    if not a_separable:
        return iter((a, i) for i in b)
    return iter((i, b) for i in a)


_KT = typing.TypeVar('_KT', bound=str)

_VT = typing.TypeVar('_VT')


class InjectiveTypeError(TypeError):
    """The given mapping contains repeated values."""


class NonInvertibleError(TypeError):
    """The given mapping is not invertible."""


class Bijection(collections.abc.Mapping, typing.Generic[_KT, _VT]):
    """An invertable mapping."""

    def __init__(self, __mapping: typing.Mapping[_KT, _VT]) -> None:
        mapping = dict(__mapping)
        n_keys = len(mapping.keys())
        n_values = len(set(mapping.values()))
        if n_keys > n_values:
            raise InjectiveTypeError(
                f"Cannot initialize {Bijection} from mapping"
                " that is injective but not surjective"
            ) from None
        if n_keys != n_values:
            raise NonInvertibleError(
                f"Cannot initialize {Bijection} from non-invertable mapping"
                f" with {n_keys} keys and {n_values} values"
            ) from None
        self._mapping = mapping

    def __iter__(self) -> typing.Iterator[str]:
        """Called for iter(self)."""
        return iter(self._mapping)

    def __len__(self) -> int:
        """Called for len(self)."""
        return len(self._mapping)

    def __getitem__(self, key):
        """Look up item by key."""
        return self._mapping[key]

    def invert(self):
        """Invert this mapping."""
        return type(self)({v: k for k, v in self.items()})

    D = typing.TypeVar('D')

    @typing.overload
    def find(self, value: _VT, /) -> _KT: ...

    @typing.overload
    def find(
        self,
        value: _VT,
        default: D,
        /,
    ) -> typing.Union[_KT, D]: ...

    def find(self, value, *args):
        """Get the unique key associated with `value`.

        This method behaves like `dict.pop`, rather than like `dict.get`, in
        that it raises a `KeyError` if `value` is not in this mapping and the
        caller did not provide `default`.

        Parameters
        ----------
        value
            The value of the desired key.
        default : optional
            A value to return if this mapping does not contain `value`.

        Returns
        ------
        key-type
            The key associated with `value`, if `value` exists.
        """
        inverted = self.invert()
        if not args:
            # The caller did not pass a default value.
            return inverted[value]
        # Extract the default value from positional arguments.
        default, *extra = args
        if not extra:
            return inverted.get(value, default)
        # There must not be any extra positional arguments.
        argstr = etc.join([str(x) for x in extra])
        s = '' if len(argstr) == 1 else '(s)'
        raise TypeError(
            f"Got extra positional argument{s} {argstr}"
        ) from None


def length(this):
    """Non-erroring version of `len(this)`.
    
    This function is intended for distinguishing between objects that have
    defined length > 1 and those that do not. It will not distinguish between
    empty iterable objects (e.g., `[]`), which have zero length, and unsized
    objects (e.g., `None`), which have undefined length. Such a distinction may
    be useful when iteratively determining whether or not to subscript a
    sequence or to convert a possibly scalar object into a sequence.
    """
    try:
        n = len(this)
    except TypeError:
        n = 0
    return n


class ObjectRegistry(collections.abc.Mapping, typing.Mapping[_KT, _VT]):
    """A class for associating metadata with abitrary objects."""
    def __init__(
        self,
        base: typing.Mapping[_KT, _VT]=None,
        object_key: str='object',
    ) -> None:
        mapping = dict(base or {})
        self._items = {
            k: dict(v) if isinstance(v, typing.Mapping) else {object_key: v}
            for k, v in mapping.items()
        }
        self._object_key = object_key
        self._init = self._items.copy()
        self._default_key_count = 1

    def __iter__(self) -> typing.Iterator[str]:
        return iter(self._items)

    def __len__(self) -> int:
        return len(self._items)

    def __contains__(self, key: str) -> bool:
        return key in self._items

    _OT = typing.TypeVar('_OT')

    def register(
        self,
        _obj: _OT=None,
        name: str=None,
        overwrite: bool=False,
        **metadata
    ) -> _OT:
        """Register an object and any associated metadata.

        This function exists to decorate objects. Without any arguments, it will
        log the decorated object in an internal mapping, keyed by the object's
        name. The user may optionally provide key-value pairs of metadata to
        associate with the object.

        Parameters
        ----------
        name : string
            The name to use as this object's key in the internal mapping. The
            default is `None`, which causes this class to create a unique key
            based on the defined name of the object.

        overwrite : bool, default=false
            If true and there is already an object with the key given by `name`,
            overwrite that object. This keyword has no effect if `name` is
            `None`.

        **metadata
            Arbitrary metadata to associate with the decorated object.

        Returns
        -------
        Any
            The decorated object.

        Examples
        --------
        Create an empty object registry, register functions with and without
        metadata, and reset the registry.

        >>> from goats.common.iterables import ObjectRegistry
        >>> from pprint import pprint
        >>> registry = ObjectRegistry()
        >>> @registry.register
        ... def myfunc():
        ...     pass
        ... 
        >>> pprint(registry.mapping)
        {'myfunc': {'object': <function myfunc at 0x7f6097427ee0>}}
        >>> @registry.register(units='cm/s')
        ... def vel():
        ...     pass
        ... 
        >>> pprint(registry.mapping)
        {'myfunc': {'object': <function myfunc at 0x7f6097427ee0>},
        'vel': {'object': <function vel at 0x7f60570fb280>, 'units': 'cm/s'}}
        >>> registry.reset()
        >>> pprint(registry.mapping)
        {}

        Create a registry with an existing `dict` and register a function with
        the same name as an object in the initializing `dict`.

        >>> from goats.common.iterables import ObjectRegistry
        >>> from pprint import pprint
        >>> mymap = {'this': [2, 3]}
        >>> registry = ObjectRegistry(mymap)
        >>> @registry.register
        ... def this():
        ...     pass
        ... 
        >>> pprint(registry.mapping)
        {'this': {'object': [2, 3]},
        'this_1': {'object': <function this at 0x7f46d49961f0>}}

        Create a registry with an existing `dict` and register a function with a
        defined name that is the same as an object in the initializing `dict`,
        but store it under a different name, then repeat the process but
        overwrite the first decorated function.

        >>> from goats.common.iterables import ObjectRegistry
        >>> from pprint import pprint
        >>> mymap = {'this': [2, 3]}
        >>> registry = ObjectRegistry(mymap)
        >>> @registry.register(name='that')
        ... def this():
        ...     pass
        >>> pprint(registry.mapping)
        {'that': {'object': <function this at 0x7f1fab723d30>},
        'this': {'object': [2, 3]}}
        >>> @registry.register(name='that', overwrite=True)
        ... def other():
        ...     pass
        >>> pprint(registry.mapping)
        {'that': {'object': <function other at 0x7f1fab723dc0>},
        'this': {'object': [2, 3]}}
        """
        @functools.wraps(_obj)
        def decorator(obj):
            key = self._get_mapping_key(obj, user=name, overwrite=overwrite)
            self._items[key] = {self._object_key: obj, **metadata}
            return obj
        if _obj is None:
            return decorator
        return decorator(_obj)

    def _get_mapping_key(
        self,
        obj: typing.Any,
        user: str=None,
        overwrite: bool=False
    ) -> str:
        """Get an appropriate key to associate with this object."""
        available = user not in self._items or overwrite
        if user and isinstance(user, typing.Hashable) and available:
            return user
        return self._get_default_key(obj)

    def _get_default_key(self, obj: typing.Any) -> str:
        """Get a default key based on the object's name and existing keys."""
        options = ['__qualname__', '__name__']
        for option in options:
            if hasattr(obj, option):
                proposed = getattr(obj, option)
                if proposed not in self._items:
                    return proposed
                new_default_key = f"{proposed}_{self._default_key_count}"
                self._default_key_count += 1
                return new_default_key

    def reset(self) -> None:
        """Reset the internal mapping to its initial state."""
        self._items = {**self._init}

    def __getitem__(self, key: str) -> typing.Dict[str, typing.Any]:
        """Get an item from the object collection."""
        return self._items[key]

    def copy(self):
        """A shallow copy of this instance."""
        return ObjectRegistry(self._items.copy(), object_key=self._object_key)

    def __repr__(self) -> str:
        """An unambiguous representation of this object."""
        return f"{self.__class__.__qualname__}({self._items})"


_NT = typing.TypeVar('_NT', bound=numbers.Complex)


class Nearest(typing.NamedTuple):
    """The result of searching an array for a target value."""

    index: int
    value: _NT


def nearest(
    values: typing.Iterable[_NT],
    target: _NT,
    bound: str=None,
) -> Nearest:
    """Find the value in a collection nearest the target value.
    
    Parameters
    ----------
    values : iterable of numbers
        An iterable collection of numbers to compare to the target value. Must
        support conversion to a `numpy.ndarray`.

    target : number
        A single numerical value for which to search in `values`. Must be
        coercible to the type of `values`.

    bound : {None, 'lower', 'upper'}
        The constraint to apply when finding the nearest value:

        - None: no constraint
        - 'lower': ensure that the nearest value is equal to or greater than the
          target value (in other words, the target value is a lower bound for
          the nearest value)
        - 'upper': ensure that the nearest value is equal to or less than the
          target value (in other words, the target value is an upper bound for
          the nearest value)

    Returns
    -------
    Nearest
        A named tuple with `value` and `index` fields, respectively containing
        the value in `values` closest to `target` (given the constraint set by
        `bound`, if any) and the index of `value` in `values`. If the array
        corresponding to `values` is one-dimensional, `index` will be an
        integer; otherwise, it will be a tuple with one entry for each
        dimension.

    Notes
    -----
    This function is based on the top answer to this StackOverflow question:
    https://stackoverflow.com/questions/2566412/find-nearest-value-in-numpy-array
    However, a lower-voted answer (and the comments) has some suggestions for a
    bisection-based method.
    """

    array = numpy.asarray(values)
    index = numpy.abs(array - target).argmin()
    if bound == 'lower':
        try:
            while array[index] < target:
                index += 1
        except IndexError:
            index = -1
    elif bound == 'upper':
        try:
            while array[index] > target:
                index -= 1
        except IndexError:
            index = 0
    if array.ndim > 1:
        index = numpy.unravel_index(index, array.shape)
    return Nearest(index=index, value=array[index])


class MergeError(Exception):
    """An error occurred while merging iterable objects."""


def merge(these: typing.Iterable, those: typing.Iterable):
    """Merge two iterable containers while respecting order.
    
    Parameters
    ----------
    these, those
        The iterable containers to merge.

    Returns
    -------
    `list`
        A list containing the unique members of the arguments, in the order in
        which they would appear after expanding both arguments.

    Raises
    ------
    `ValueError`
        The arguments contain repeated items in different order.
    """
    x = list(these)
    y = list(those)
    repeated = set(x) & set(y)
    if repeated:
        ab = [i for i in x if i in y]
        ba = [i for i in y if i in x]
        if ab != ba:
            raise MergeError(
                "Repeated entries must appear in the same order"
            ) from None
        s = []
        za = zb = 0
        for v in ab:
            ia = x.index(v)
            ib = y.index(v)
            s.extend(x[za:ia] + y[zb:ib])
            s.append(v)
            za = ia+1
            zb = ib+1
        s.extend(x[za:] + y[zb:])
        return s
    x.extend(y)
    return x


def isiterable(x, /):
    """True if the argument is iterable."""
    try:
        iter(x)
    except TypeError:
        return False
    return True


def size(x, /) -> int:
    """Compute the size of a potentially nested collection.
    
    Parameters
    ----------
    x
        Any non-string iterable container.

    Notes
    -----
    - The non-string restriction on `x` exists to restrict this function to
      array-like objects (e.g., lists of lists). Such objects are considered
      "separable".

    Raises
    ------
    TypeError
        `x` is not iterable or is not separable

    See Also
    --------
    `~isseparable`
    """
    if not isiterable(x):
        raise TypeError(
            f"Cannot compute the size of non-iterable {x}"
        ) from None
    if not isseparable(x):
        raise TypeError(
            f"Argument must be a separable collection, not {type(x)}"
        ) from None
    count = 0
    for y in x:
        if isiterable(y):
            count += size(y)
        else:
            count += 1
    return count


def slice2range(s: slice, /, stop: int=None) -> range:
    """Attempt to convert a `slice` to the equivalent `range`.
    
    Parameters
    ----------
    s : slice
        The `slice` object to convert.

    stop : int, optional
        If given, this function will use the value of `stop` as the (required)
        stop value of the resultant `range` object if the stop value of the
        given `slice` is `None`.

    Returns
    -------
    `range`
        A `range` object built from appropriate start, stop, and step values. If
        the given `slice` doesn't define a start value, this function will use
        0. If the given `slice` doesn't define a step value, this function will
        use 1. See Parameters and Raises for descriptions of behavior for
        different stop values.

    Raises
    ------
    `TypeError`
        Both the stop value of the given `slice` and the value of the `stop`
        keyword parameter are `None`. It is not possible to create a `range`
        object with non-integral stop value.
    """
    if stop is None and s.stop is None:
        raise TypeError(f"Cannot convert {s} to a range.")
    start = s.start or 0 # same effect if s.start == 0
    stop = stop if s.stop is None else s.stop
    step = s.step or 1 # handles invalid case of s.step == 0
    return range(start, stop, step)


def isshapelike(x, /):
    """True if `x` looks like an array shape.
    
    This function will return `True` if `x` is a non-empty iterable object that
    contains only integral members. Otherwise, it will return `False`.
    """
    if not x:
        return False
    return (
        isiterable(x)
        and
        all(isinstance(i, numbers.Integral) for i in x)
    )





import typing

from ._expression import (
    Expression,
    expression_factory as expression,
    reduce,
    standard,
)
from ._operand import (
    asterms,
    OperandFactory,
    OperandTypeError,
    OperandValueError,
    Term,
    term_factory as term,
)
from ._parser import (
    ParsingValueError,
    ProductError,
    RatioError,
)

__all__ = [
    'Expression',
    'Expressable',
    'OperandFactory',
    'OperandTypeError',
    'OperandValueError',
    'ParsingValueError',
    'ProductError',
    'RatioError',
    'Term',
    'asterms',
    'expression',
    'reduce',
    'standard',
    'term',
]


Expressable = typing.TypeVar(
    'Expressable',
    str,
    typing.Iterable,
    Expression,
)
Expressable = typing.Union[str, typing.Iterable, Expression]


def composition(this):
    """True if `this` appears to be a symbolic composition of terms.
    
    Parameters
    ----------
    this
        The object to check.

    Notes
    -----
    This is more stringent than simply checking whether `this` can instantiate a
    `~symbolic.Expression` because any string would satisfy that condition.
    """
    return (
        isinstance(this, Expression)
        or isinstance(this, str) and ('/' in this or '*' in this)
    )


def equality(a, b) -> Expression:
    """Symbolically compute a == b."""
    x, y = (expression(i) for i in (a, b))
    return x == y


def product(a, b) -> Expression:
    """Symbolically compute a * b."""
    x, y = (expression(i) for i in (a, b))
    return x * y


def ratio(a, b) -> Expression:
    """Symbolically compute a / b."""
    x, y = (expression(i) for i in (a, b))
    return x / y


def power(a, n) -> Expression:
    """Symbolically compute a ** n."""
    return expression(a) ** n



"""
Partial implementations of certain abstract protocols, built on the
corresponding abstract base class.
"""

import typing

from . import _abcs
from . import _protocols


class Comparable(_abcs.Comparable):
    """Partial implementation of the comparable protocol.
    
    This mixin class provides concrete definitions of the following methods,
    based on concrete definitions of `__eq__` (called for `a == b`) and `__lt__`
    (called for `a < b`)

    - `__ne__`: not `a == b`
    - `__le__`: `a < b` or `a == b`
    - `__gt__`: not `a < b` and not `a == b`
    - `__ge__`: not `a < b`

    See Also
    --------
    `~quantity.Comparable`
        The corresponding abstract protocol class.
    `~quantity.abc.Comparable`
        The corresponding abstract base class.
    """

    def __ne__(self: _protocols.Comparable, other):
        return not (self == other)

    def __le__(self: _protocols.Comparable, other):
        return (self < other) or (self == other)

    def __gt__(self: _protocols.Comparable, other):
        return not (self < other) and not (self == other)

    def __ge__(self: _protocols.Comparable, other):
        return not (self < other)


class Complex(_abcs.Complex):
    """Partial implementation of the complex-valued protocol.
    
    This mixin class provides concrete definitions of the following methods,
    both based on concrete definitions of `__add__` (called for `a + b`) and
    `__neg__` (called for `-a`)

    - `__sub__`:  `a + (-b)`
    - `__rsub__`: `b + (-a)`

    See Also
    --------
    `~quantity.Complex`
        The corresponding abstract protocol class.
    `~quantity.abc.Complex`
        The corresponding abstract base class.
    """

    def __sub__(self: _protocols.Complex, other):
        if isinstance(other, _protocols.Complex):
            return self + (-other)
        return NotImplemented

    def __rsub__(self: _protocols.Complex, other):
        if isinstance(other, _protocols.Complex):
            return other + (-self)
        return NotImplemented


class Scalar(_abcs.Scalar):
    """Partial implementation of the scalar protocol.
    
    This mixin class provides a concrete definition of `__complex__`, based on a
    concrete definition of `__float__`.

    See Also
    --------
    `~quantity.Scalar`
        The corresponding abstract protocol class.
    `~quantity.abc.Scalar`
        The corresponding abstract base class.
    """

    def __complex__(self: _protocols.Scalar) -> complex:
        return complex(float(self))


class Sequence(_abcs.Sequence):
    """Partial implementation of the sequence-like protocol.
    
    This mixin class provides concrete definitions of the following methods,
    based on concrete definitions of `__getitem__` and `__iter__`. Both
    definitions mimic the respective definitions on `collections.abc.Sequence`.

    - `__contains__`
    - `__iter__`

    See Also
    --------
    `~quantity.Sequence`
        The corresponding abstract protocol class.
    `~quantity.abc.Sequence`
        The corresponding abstract base class.
    """

    def __contains__(self: _protocols.Sequence, v, /) -> bool:
        """Called for v in self."""
        for i in self:
            if i is v or v == i:
                return True
        return False

    def __iter__(self: _protocols.Sequence) -> typing.Iterator:
        """Called for iter(self)."""
        i = 0
        try:
            while True:
                v = self[i]
                yield v
                i += 1
        except IndexError:
            return




"""
Initialization contexts for measured objects.
"""

import typing

from .. import etc
from .. import metric


T = typing.TypeVar('T')

Converter = typing.Callable[[metric.Conversion, T], T]


@etc.autostr
class Context(typing.Generic[T]):
    """Concrete context for objects with a metric unit."""

    def __init__(
        self,
        unit: metric.UnitLike,
        converter: Converter[T],
        /,
    ) -> None:
        self._unit = unit
        self._converter = converter

    def convert(self, x: T, new: metric.UnitLike) -> T:
        """Convert the given object to a new unit.

        Parameters
        ----------
        x : `~T`
            The object to convert.
        new : unit-like
            The unit to which to convert `x`.

        Returns
        -------
        `~T`
            The original object or an object with the same type.

        Raises
        ------
        `ValueError`
            The proposed unit is inconsistent with the given object's current
            unit. Two units are consistent if they have the same dimension in a
            known metric system.
        """
        if metric.unitlike(new):
            try:
                c = metric.conversion(self.unit, new)
            except metric.UnitConversionError as err:
                if not (self.unit | new):
                    raise ValueError(
                        f"The unit {str(new)!r} is inconsistent"
                        f" with {str(self.unit)!r}"
                    ) from err
                raise err
            return self.apply(c, x)
        raise TypeError(
            f"Cannot interpret {new!r} as a unit"
        ) from None

    @property
    def unit(self):
        """The metric unit associated with this context."""
        return self._unit

    def apply(self, c: metric.Conversion, x: T) -> T:
        """Apply the given unit conversion to the given object."""
        return self.converter(c, x)

    @property
    def converter(self):
        """The unit-conversion application."""
        return self._converter


@typing.overload
def context_factory(
    unit: metric.UnitLike=typing.Literal['1'],
    converter: typing.Optional[Converter[T]]=None,
) -> Context[T]: ...

def context_factory(*args, **kwargs):
    """Factory function for `~Context`."""
    try:
        unit, converter = context_args(*args, **kwargs)
    except (TypeError, ValueError) as err:
        argstr = ', '.join(str(arg) for arg in args)
        kwdstr = ', '.join(f"{k}={v}" for k, v in kwargs.items())
        callstr = f"{argstr}, {kwdstr}" if kwdstr else argstr
        raise TypeError(
            f"Failed to initialize a context from ({callstr})"
        ) from err
    return Context(unit, converter)


def context_args(*args, **kwargs) -> typing.Tuple[metric.Unit, Converter[T]]:
    """Parse arguments to initialize an instance of `~Context`."""
    if args and not kwargs:
        return metric.unit(args[0] or '1'), args[1] or _default_converter
    if kwargs and not args:
        return (
            metric.unit(kwargs.get('unit') or '1'),
            kwargs.get('converter') or _default_converter,
        )
    if len(args) == 1:
        return (
            metric.unit(args[0] or '1'),
            kwargs.get('converter') or _default_converter,
        )
    message = f"{args}, {', '.join(f'{k}={v}' for k, v in kwargs.items())}"
    raise TypeError(message)


def _default_converter(c: metric.Conversion, x: T) -> T:
    """Default scaling function for unit conversion."""
    return type(x)(float(c) * x)





"""
Utilities for defining factories of measured objects.
"""

import functools
import inspect
import typing

from ._objects import (
    Object,
    ObjectT,
)


def strict(*args):
    """Decorator function for enforcing attribute immutability.

    This function will raise a `ValueError` if it detects an attempt change the
    unit when creating a measured object from an existing measured object via
    the decorated factory function.
    """
    if all(isinstance(arg, str) for arg in args):
        func, keys = None, args
    elif len(args) == 1:
        arg = args[0]
        if callable(arg):
            func, keys = arg, ()
    else:
        raise TypeError(args)
    def decorate(__callable: typing.Callable[..., ObjectT]):
        return _check_metadata(__callable, *keys)
    if func is None:
        return decorate
    return decorate(func)


def returns(__type: typing.Type[ObjectT], strict: bool=True):
    """Mark the decorated object as the factory for the given type.

    Parameters
    ----------
    __type : subclass of `~Object`
        The type of object that the decorated function creates.

    strict : bool, default=True
        By default, this function will require that a given factory does not
        support changing the unit when creating a measured object from an
        existing measured object. Setting `strict=False` removes this
        restriction.

    See Also
    --------
    `~Object.factory`
        A class-method decorator that provides similar functionality.
    """
    def decorated(__callable: typing.Callable[..., ObjectT]):
        if issubclass(__type, Object):
            if strict:
                return __type.register.factory(_check_metadata(__callable))
            return __type.register.factory(__callable)
        raise TypeError(
            f"Cannot use {__callable} as the factory for {__type}"
            f", which is not a subtype of {Object}"
        ) from None
    return decorated


def _check_metadata(f: typing.Callable[..., ObjectT], *keys: str):
    """Internal logic for enforcing metadata immutability in factories."""
    rettype = inspect.signature(f).return_annotation
    newtype = Object if rettype is inspect.Signature.empty else rettype
    @functools.wraps(f)
    def wrapped(x, /, **kwargs):
        rtstr = newtype.__qualname__
        xtstr = type(x).__qualname__
        rest = (
            f"when creating an instance of {rtstr}"
            f" from an existing instance of {xtstr}"
        )
        errmsg = "Cannot change {} " + rest
        if isinstance(x, Object):
            if not keys:
                for k, v in kwargs.items():
                    if v is not None:
                        raise ValueError(errmsg.format(k)) from None
            for key in keys:
                if kwargs.get(key) is not None:
                    raise ValueError(errmsg.format(key)) from None
        return f(x, **kwargs)
    return wrapped





"""
Support for working with metric systems.
"""

import collections.abc
import typing

from .. import etc
from .. import symbolic
from . import _conversions
from . import _defined
from . import _dimensions
from . import _exceptions
from . import _reference
from . import _units


@etc.autostr
class Properties:
    """Canonical properties of a quantity within a metric system."""

    def __init__(
        self,
        system: str,
        unit: typing.Union[str, _units.Unit],
    ) -> None:
        self._system = system.lower()
        self._unit = _units.unit_factory(unit)
        self._dimension = None

    @property
    def unit(self):
        """The canonical unit of this quantity in this metric system."""
        return self._unit

    @property
    def dimension(self):
        """The dimension of this quantity in this metric system."""
        if self._dimension is None:
            self._dimension = self.unit.dimensions[self._system]
        return self._dimension

    def __eq__(self, __o) -> bool:
        """True if two instances have equal units and dimensions."""
        try:
            equal = [
                getattr(self, attr) == getattr(__o, attr)
                for attr in ('unit', 'dimension')
            ]
            return all(equal)
        except AttributeError:
            return False

    def __str__(self) -> str:
        properties = ', '.join(
            f"{p}={str(getattr(self, p, None))!r}"
            for p in ['unit', 'dimension']
        )
        return f"{properties} [{self._system!r}]"


@etc.autostr
class Quantity:
    """A single metric quantity.

    An instance of this class represents the properties of the named metric
    quantity in all defined metric systems.
    """

    def __init__(self, name: str) -> None:
        self._name = name
        self._units = None
        self._dimensions = None

    def __getitem__(self, system: str):
        """Get this quantity's representation in the named metric system."""
        try:
            unit = self.units[system.lower()]
            return Properties(system, unit)
        except KeyError as err:
            raise _exceptions.QuantityError(
                f"No properties available for {self.name!r} in {system!r}"
            ) from err

    def convert(self, unit: str) -> _conversions.Converter:
        """Create a conversion handler for `unit`.

        Parameters
        ----------
        unit : string
            The unit of this quantity from which to convert.

        Returns
        -------
        `~Converter`
            An instance of the unit-conversion handler. The returned object
            supports conversion from `unit` to any unit for which the conversion
            is defined, or to the equivalent unit in a given metric system.

        Notes
        -----
        - This method is defined within the scope of a particular metric
          quantity because unit conversions are only defined relative to their
          respective quantities, even though two quantities may have identical
          conversions (e.g., frequency and vorticity). The distinction between
          quantities is necessary when dealing with CGS electromagnetic units.
          For example, 'cm / s' is the canonical unit of both conductance and
          velocity in CGS; converting to its equivalent MKS unit therefore
          requires knowledge of the relevant quantity.
        """
        return _conversions.converter_factory(unit, self.name)

    _attrs = ('dimensions', 'units')

    def __eq__(self, other) -> bool:
        """Called to test equality via self == other.
        
        Two instances of this class are either identical, in which case they are
        triviall equal, or they represent distinct quantities, in which they are
        not equal. In addition, an instance of this class will compare equal to
        its case-insensitive name.

        See Also
        --------
        `~metric.Quantity.__or__` : test equivalence between quantities.
        """
        if isinstance(other, str):
            return other.lower() == self.name
        return other is self

    def __or__(self, other) -> bool:
        """Called to test equivalence via self | other.
        
        Two metric quantities are equivalent if their dimensions are equal in a
        given metric system. This operation thus provides a way to compare
        unequal quantities to determine if they are have a physical
        correspondence. For example, energy and work are distinct quantities,
        but they have identical dimensions and are linked through the
        work-energy theorem.

        This operation is only defined between two instances of this class.
        """
        if isinstance(other, Quantity):
            for system in _reference.SYSTEMS:
                if self[system].dimension != other[system].dimension:
                    return False
            return True
        return NotImplemented

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return self.name

    @property
    def units(self):
        """The unit of this quantity in each metric system."""
        if self._units is None:
            self._units = _defined.UNITS[self.name]
        return self._units

    @property
    def dimensions(self):
        """The dimension of this quantity in each metric system."""
        if self._dimensions is None:
            self._dimensions = _defined.DIMENSIONS[self.name]
        return self._dimensions

    @property
    def name(self):
        """The name of this physical quantity."""
        return self._name


_quantities = {}
"""Internal collection of singleton `~Quantity` instances."""

def quantity_factory(arg: typing.Union[str, Quantity]) -> Quantity:
    """Factory function for metric quantities.

    Parameters
    ----------
    arg : string or `~Quantity`.
        The metric quantity to represent. If `arg` is already an instance of
        `~Quantity`, this function will immediately return it.

    Returns
    -------
    `~Quantity`
    """
    if isinstance(arg, Quantity):
        return arg
    name = str(arg).lower()
    if available := _quantities.get(name):
        return available
    q = Quantity(name)
    _quantities[name] = q
    return q


class SearchError(KeyError):
    """Error while searching for a requested metric."""
    pass


@etc.autostr
class System(collections.abc.Mapping):
    """Representations of physical quantities within a given metric system."""

    def __init__(self, name: str) -> None:
        self._name = name
        self._units = None
        self._dimensions = None

    def __len__(self) -> int:
        """The number of quantities defined in this metric system."""
        return _reference._QUANTITIES.__len__()

    def __iter__(self) -> typing.Iterator:
        """Iterate over defined metric quantities."""
        return _reference._QUANTITIES.__iter__()

    def __getitem__(self, key: typing.Union[str, Quantity]):
        """Get the metric for the requested quantity in this system."""
        try:
            found = quantity_factory(key)
        except ValueError as err:
            raise KeyError(f"No known quantity called '{key}'") from err
        else:
            return found[self.name]

    def get_dimension(self, this: typing.Union[str, _units.Unit]):
        """Compute the dimension of `unit` in this metric system."""
        unit = _units.unit_factory(this)
        expression = symbolic.expression('1')
        systems = set()
        for term in unit:
            named = _defined.NamedUnit(term.base)
            allowed = named.systems['allowed']
            dimension = (
                named.dimensions[self.name] if len(allowed) > 1
                else named.dimensions[allowed[0]]
            )
            expression *= symbolic.expression(dimension) ** term.exponent
            systems.update(allowed)
        if self.name in systems:
            return _dimensions.dimension_factory(expression)
        raise ValueError(
            f"Can't define dimension of {unit!r} in {self.name!r}"
        ) from None

    def get_unit(
        self,
        *,
        unit: typing.Union[str, _units.Unit]=None,
        dimension: typing.Union[str, _dimensions.Dimension]=None,
        quantity: typing.Union[str, Quantity]=None,
    ) -> typing.Optional[_units.Unit]:
        """Get a canonical unit from a given unit, dimension, or quantity.

        This method will search for the unit in the current metric system based
        on `unit`, `dimension`, or `quantity`, in that order. All arguments
        default to `None`. If `unit` is not `None`, this method will attempt to
        return the equivalent canonical unit; if either `dimension` or
        `quantity` is not `None`, this method will attempt to return the unique
        corresponding unit.

        Parameters
        ----------
        unit : string or Unit
            A unit of measure in any system.

        dimension : string or Dimension
            A physical dimension.

        quantity : string or Quantity
            A physical quantity.

        Returns
        -------
        Unit
            The corresponding unit in the current metric system.

        """
        methods = {
            k: getattr(self, f'_unit_from_{k}')
            for k in ('unit', 'dimension', 'quantity')
        }
        targets = {
            'unit': unit,
            'dimension': dimension,
            'quantity': quantity,
        }
        return self._get_unit(methods, targets)

    T = typing.TypeVar('T', str, _units.Unit, _dimensions.Dimension, Quantity)
    T = typing.Union[str, _units.Unit, _dimensions.Dimension, Quantity]

    def _get_unit(
        self,
        methods: typing.Dict[str, typing.Callable[[T], _units.Unit]],
        targets: typing.Dict[str, T],
    ) -> _units.Unit:
        """Search logic for `get_unit`."""
        nonnull = {k: v for k, v in targets.items() if v}
        cases = [(methods[k], v) for k, v in nonnull.items()]
        for (method, arg) in cases:
            if str(arg) == '1':
                return _units.unit_factory(self['identity'].unit)
            if result := method(arg):
                return result
        args = self._format_targets(nonnull)
        errmsg = f"Could not determine unit in {self.name} from {args}"
        raise SearchError(errmsg)

    def _format_targets(self, targets: typing.Dict[str, T]):
        """Format `get_unit` targets for printing."""
        if not targets:
            return "nothing"
        args = [f'{k}={v!r}' for k, v in targets.items()]
        if 0 < len(args) < 3:
            return ' or '.join(args)
        return f"{', '.join(str(arg) for arg in args[:-1])}, or {args[-1]}"

    def _unit_from_unit(
        self,
        target: typing.Union[str, _units.Unit],
    ) -> _units.Unit:
        """Get the canonical unit corresponding to the given unit."""
        unit = _units.unit_factory(target)
        return unit.normalize(self.name)

    def _unit_from_dimension(
        self,
        target: typing.Union[str, _dimensions.Dimension],
    ) -> _units.Unit:
        """Get the canonical unit corresponding to the given dimension."""
        for quantity, dimension in self.dimensions.items():
            if dimension == target:
                return _units.unit_factory(self.units[quantity])

    def _unit_from_quantity(
        self,
        quantity: typing.Union[str, Quantity],
    ) -> _units.Unit:
        """Get the canonical unit corresponding to the given quantity."""
        return _units.unit_factory(self[quantity].unit)

    def __eq__(self, other) -> bool:
        """True if two systems have the same `name` attribute."""
        if isinstance(other, System):
            return self.name == other.name.lower()
        if isinstance(other, str):
            return self.name == other.lower()
        return NotImplemented

    def __bool__(self) -> bool:
        """A defined metric system is always truthy."""
        return True

    def keys(self) -> typing.KeysView[str]:
        return super().keys()

    def values(self) -> typing.ValuesView[Properties]:
        return super().values()

    def items(self) -> typing.ItemsView[str, Properties]:
        return super().items()

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return str(self.name)

    @property
    def units(self):
        """The unit of each physical quantity in the metric system."""
        if self._units is None:
            self._units = _defined.CANONICAL['units'][self.name]
        return self._units

    @property
    def dimensions(self):
        """The dimension of each physical quantity in the metric system."""
        if self._dimensions is None:
            self._dimensions = _defined.CANONICAL['dimensions'][self.name]
        return self._dimensions

    @property
    def name(self):
        """The name of this metric system."""
        return self._name


_systems = {}
"""Internal collection of singleton `~System` instances."""

def system_factory(arg: typing.Union[str, System]) -> System:
    """Factory function for metric-system representations.

    Parameters
    ----------
    arg : string or `~System`.
        The metric system to represent. If `arg` is already an instance of
        `~System`, this function will immediately return it.

    Returns
    -------
    `~System`
    """
    if isinstance(arg, System):
        return arg
    name = str(arg).lower()
    if available := _systems.get(name):
        return available
    s = System(name)
    _systems[name] = s
    return s





"""
Tools and scripts for analyzing EPREM simulation runs.
"""

# read version from installed package
from importlib.metadata import version
__version__ = version("eprempy")

from .physical import Array
from .observable import Quantity as Observable


__all__ = [
    "Array",
    "Observable",
]




"""
Tools for viewing the contents of an EPREM datafile.

For the purposes of this package, a datafile is a file containing simulation
output for a single observer. Each EPREM dataset contains multiple datafiles.
"""

import abc
import pathlib
import typing

import numpy.typing
import netCDF4

from .. import aliased
from .. import etc
from .. import exceptions
from .. import paths
from ._core import (
    Array,
    Scalar,
    Axis,
    array_factory,
    scalar_factory,
    axis_factory,
)
from ._reference import (
    ARRAYS,
    SCALARS,
    AXES,
)


T = typing.TypeVar('T')

class Viewer(abc.ABC):
    """An abstract base class for datafile viewers."""

    def __init__(self, source: pathlib.Path) -> None:
        self._source = source

    @abc.abstractmethod
    def get_array(self, name: str) -> Array:
        """Get the named array-like object from this datafile."""

    @abc.abstractmethod
    def get_scalar(self, name: str) -> Scalar:
        """Get the named scalar object from this datafile."""

    @abc.abstractmethod
    def get_axis(self, name: str) -> Axis:
        """Get the named array axis from this datafile."""

    @abc.abstractmethod
    def get_unit(self, name: str) -> str:
        """Get the unit for the named datafile object."""

    @property
    def source(self):
        """This datafile's source."""
        return self._source


class NetCDFViewer(Viewer):
    """A viewer for NetCDF datafiles."""

    def __init__(self, source: pathlib.Path) -> None:
        super().__init__(source)
        datafile = netCDF4.Dataset(self.source, 'r')
        self._variables = getattr(datafile, 'variables')
        self._dimensions = getattr(datafile, 'dimensions')

    def get_array(self, name: str):
        data = self._get_variable(name)
        unit = self.get_unit(name)
        dimensions = self._get_dimensions_from_data(data)
        return array_factory(data=data, unit=unit, dimensions=dimensions)

    def get_scalar(self, name: str):
        data = self._get_variable(name)
        if shape := getattr(data, 'shape', None):
            raise TypeError(
                f"Cannot convert {name!r}, with shape {shape}, to a scalar"
            ) from None
        dtype = getattr(data, 'dtype', numpy.dtype(numpy.float64))
        value = (
            int(data[:]) if numpy.issubdtype(dtype, numpy.integer)
            else float(data[:])
        )
        unit = self.get_unit(name)
        return scalar_factory(value=value, unit=unit)

    def get_axis(self, name: str):
        try:
            data = self._dimensions[name]
        except KeyError as err:
            raise KeyError(
                f"No dimension named {name!r}"
                f" in the datafile at {self.source}"
            ) from err
        size = getattr(data, 'size', None)
        return axis_factory(size=size)

    def get_unit(self, name: str) -> str:
        data = self._get_variable(name)
        return self._get_unit_from_data(data)

    def _get_variable(self, name: str):
        """Helper for retrieving datafile variables by name."""
        try:
            data = self._variables[name]
        except KeyError as err:
            raise KeyError(
                f"No variable named {name!r}"
                f" in the datafile at {self.source}"
            ) from err
        return data

    def _get_dimensions_from_data(self, data):
        """Compute appropriate variable dimensions from a datafile object."""
        return tuple(getattr(data, 'dimensions', ()))

    def _get_unit_from_data(self, data):
        """Compute appropriate variable units from a datafile object."""
        available = (
            getattr(data, attr) for attr in ('unit', 'units')
            if hasattr(data, attr)
        )
        return next(available, None)


HINTS = {
    '.nc': ['netcdf', 'NetCDF', 'NETCDF'],
}
"""Non-suffix aliases for datafile file types."""


_viewers = {
    '.nc': NetCDFViewer,
}
VIEWERS = aliased.Mapping(_viewers, aliases=HINTS)
"""The available datafile-viewing classes."""


@etc.autostr
class View(metaclass=paths.PathSet):
    """A format-agnostic view of a datafile.

    An instance of this class provides access to variables and axes defined in a
    specific datafile, given a path to that datafile. It is designed to provide
    a single interface, regardless of file type, with as little overhead as
    possible. Therefore, it does not attempt to modify attributes (e.g.,
    converting variable units), since doing so could result in reading a
    potentially large array from disk.
    """

    def __init__(
        self,
        source: pathlib.Path,
        hint: typing.Optional[str],
    ) -> None:
        self._source = source
        self._hint = hint
        self._viewer = None
        self._arrays = None
        self._scalars = None
        self._sizes = None
        self._axes = None
        self._units = None

    def __str__(self) -> str:
        return str(self.source)

    @typing.overload
    def __getitem__(
        self,
        key: typing.Literal['arrays'],
    ) -> typing.Dict[str, Array]: ...

    @typing.overload
    def __getitem__(
        self,
        key: typing.Literal['scalars'],
    ) -> typing.Dict[str, Scalar]: ...

    @typing.overload
    def __getitem__(
        self,
        key: typing.Literal['axes'],
    ) -> typing.Dict[str, Axis]: ...

    def __getitem__(self, __k):
        """Access properties by name."""
        try:
            this = getattr(self, __k)
        except AttributeError as err:
            raise KeyError(
                f"{self.__class__} has no property named {__k!r}"
            ) from err
        return this

    @property
    def units(self):
        """The unit of each datafile quantity, if available."""
        if self._units is None:
            reference = list(ARRAYS) + list(SCALARS)
            self._units = self._map(self.viewer.get_unit, reference)
        return self._units

    @property
    def arrays(self):
        """The array-like quantities in this datafile."""
        if self._arrays is None:
            self._arrays = self._map(self.viewer.get_array, ARRAYS)
        return self._arrays

    @property
    def scalars(self):
        """The scalar quantities in this datafile."""
        if self._scalars is None:
            self._scalars = self._map(self.viewer.get_scalar, SCALARS)
        return self._scalars

    @property
    def sizes(self):
        """The length of each datafile dimension."""
        if self._sizes is None:
            self._sizes = {k: v.size for k, v in self.axes.items()}
        return self._sizes

    @property
    def axes(self):
        """The array-axis quantities in this datafile."""
        if self._axes is None:
            self._axes = self._map(self.viewer.get_axis, AXES)
        return self._axes

    def _map(
        self,
        get: typing.Callable[[str], T],
        reference: typing.Iterable[str],
    ) -> typing.Dict[str, T]:
        """Create a mapping of datafile attributes."""
        guarded = exceptions.Wrapper(get)
        guarded.catch(KeyError)
        mapped = {k: guarded.compute(k) for k in reference}
        return {k: v for k, v in mapped.items() if v is not None}

    @property
    def viewer(self) -> Viewer:
        """The appropriate file viewer for this datafile."""
        if self._viewer is None:
            try:
                viewer = VIEWERS[self.source.suffix]
            except KeyError:
                viewer = VIEWERS.get(self._hint)
            self._viewer = viewer(self.source)
        return self._viewer

    @property
    def source(self):
        """The file containing output data."""
        return self._source


def view_factory(
    source: paths.PathLike,
    *,
    hint: typing.Optional[str]=None,
) -> View:
    """Create a format-agnostic view of an EPREM datafile.

    Parameters
    ----------
    source : string or `os.PathLike`
        The path to the target datafile. The path must exist. This function will
        convert `source` into a fully qualified `pathlib.Path`.
    hint : optional; keyword only
        A defined alias for the target file type. See Notes for further
        explanation.

    Notes
    -----
    - The datafile-viewing interface will attempt to determine the correct
      file-type viewer for the target datafile based on the path suffix.
      Providing a `hint` allows the interface to choose the appropriate viewer
      even in cases when the target file does not have the expected suffix.
    """
    return View(paths.fullpath(source, strict=True), hint=hint)





import abc
import typing

from .. import base
from .. import numeric
from .. import metric


T = typing.TypeVar('T')


class Interface(numeric.Interface, typing.Protocol[T]):
    """Abstract protocol class for measured objects.

    A measured object has real-valued data and an associated metric unit.

    Objects that implement the numeric and metric protocols, and that define the
    following method, are compatible with this protocol

    - `to`, which should return a new instance of the implementing class
    """

    @property
    @abc.abstractmethod
    def isunitless(self: T) -> bool:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def unit(self: T) -> metric.Unit:
        raise NotImplementedError

    @abc.abstractmethod
    def withunit(self: T, arg: metric.UnitLike) -> T:
        raise NotImplementedError


class Object(numeric.Object[base.RealType], Interface):
    """Abstract base class for measured objects.

    A measured object has real-valued data and an associated metric unit.

    This class adds an internal `_unit` attribute to `~numeric.Object`. It
    defines `isunitless` and `unit` properties. Concrete subclasses must
    implement the abstract method `to` from `~measured.Interface`.
    """

    @numeric.data.implements(base.Real)
    def __init__(
        self,
        data: base.RealType,
        unit: metric.Unit,
        /,
    ) -> None:
        super().__init__(data)
        self._unit = unit

    def __str__(self) -> str:
        return f"{self.data}, unit={str(self.unit)!r}"

    def __eq__(self, other) -> bool:
        if isinstance(other, Object):
            return (
                numeric.data.isequal(self.data, other.data)
                and
                self.unit == other.unit
            )
        return False

    @property
    def isunitless(self) -> bool:
        """True if this object's unit is '1'."""
        return self.unit == '1'

    @property
    def unit(self):
        """The metric unit of this object's data."""
        return self._unit


@typing.runtime_checkable
class Type(Interface[T], typing.Protocol):
    """Structural protocol class for measured objects.

    A measured object has real-valued data and an associated metric unit.

    In addition to attributes defined on `~data.Type`, calling code can expect
    the following attributes on an object that passes `isinstance(...,
    measured.Type)`

    - `unit` (property)
    - `isunitless` (property)
    - `to` (method)
    """

    _unit: metric.Unit

    @property
    def isunitless(self: T) -> bool: ...

    @property
    def unit(self: T) -> metric.Unit: ...

    def withunit(self: T, arg: metric.UnitLike) -> T: ...





import numbers
import typing

import numpy

from .. import metric
from ._context import (
    context_factory,
)
from ._object import Object
from ._value import Value
from ._sequence import Sequence
from . import _convert


@Value.register.factory
def value_factory(x, /, unit=None) -> Value:
    """Factory function for `~Value`."""
    d, u = value_args(x, unit)
    ctx = context_factory(u, _convert.singular)
    return Value(d, ctx)


def value_args(x, unit, /) -> typing.Tuple[typing.Any, metric.Unit]:
    """Parse arguments to initialize `~Value`."""
    if isinstance(x, Value):
        return x.data, x.unit
    if isinstance(x, (int, float, numbers.Real)):
        return x, unit
    if isinstance(x, (Sequence, numpy.ndarray)) and x.size == 1:
        if isinstance(x, Sequence):
            return x.data[0], x.unit
        if isinstance(x, numpy.ndarray):
            return value_args(numpy.ravel(x)[0], unit)
    if isinstance(x, typing.Sequence) and not isinstance(x, str):
        return value_args(numpy.array(x), unit)
    raise TypeError(
        f"Cannot create a value from {x!r} and {unit!r}"
    ) from None


@Sequence.register.factory
def sequence_factory(x, /, unit=None) -> Sequence:
    """Factory function for `~Sequence`."""
    d, u = sequence_args(x, unit)
    ctx = context_factory(u, _convert.arraylike)
    return Sequence(d, ctx)


def sequence_args(x, unit, /) -> typing.Tuple[typing.Any, metric.Unit]:
    """Parse arguments to initialize `~Sequence`."""
    if isinstance(x, Value):
        return sequence_args([x.data], x.unit)
    if isinstance(x, Object):
        return x.data, x.unit
    if isinstance(x, numbers.Real):
        return numpy.array([x]), unit
    a = numpy.asarray(x)
    if a.ndim > 0:
        return a.flatten(), unit
    raise TypeError(
        f"Cannot create a sequence from {x!r} and {unit!r}"
    ) from None





"""
Names of quantities in an arbitrary EPREM dataset.
"""


ARRAYS = (
    'phiOffset',
    'time',
    'shell',
    'mu',
    'mass',
    'charge',
    'egrid',
    'vgrid',
    'R',
    'T',
    'P',
    'Br',
    'Bt',
    'Bp',
    'Vr',
    'Vt',
    'Vp',
    'Rho',
    'mfp',
    'Dist',
    'flux',
)
"""The canonical names of array-like dataset quantities.

Note that not every named quantity will appear in each dataset. For example,
datasets generated by a run with `outputFlux=1` will include 'flux' but not
'Dist', while datasets generated by a run with `outputFlux=0` will include
'Dist' but not 'flux'.
"""


SCALARS = (
    'preEruption',
)
"""The canonical names of scalar dataset quantities."""


AXES = (
    'time',
    'shell',
    'species',
    'energy',
    'mu',
)
"""The canonical names of dataset axes."""





import typing

from .. import aliased
from .. import container
from . import _operators


T = typing.TypeVar('T')


_OPERATION_METADATA = {
    'int': {
        'builtin': int,
        'forward': _operators.__int__,
    },
    'float': {
        'builtin': float,
        'forward': _operators.__float__,
    },
    'complex': {
        'builtin': complex,
        'forward': _operators.__complex__,
    },
    'eq': {
        'builtin': _operators.eq,
        'forward': _operators.__eq__,
        'aliases': [
            'equal',
        ],
    },
    'ne': {
        'builtin': _operators.ne,
        'forward': _operators.__ne__,
        'aliases': [
            'not equal',
            'unequal',
        ],
    },
    'lt': {
        'builtin': _operators.lt,
        'forward': _operators.__lt__,
        'aliases': [
            'less than',
            'less',
        ],
    },
    'le': {
        'builtin': _operators.le,
        'forward': _operators.__le__,
        'aliases': [
            'less than or equal to',
            'less or equal',
        ],
    },
    'gt': {
        'builtin': _operators.gt,
        'forward': _operators.__gt__,
        'aliases': [
            'greater than',
            'greater',
        ],
    },
    'ge': {
        'builtin': _operators.ge,
        'forward': _operators.__ge__,
        'aliases': [
            'greater than or equal to',
            'greater or equal',
        ],
    },
    'abs': {
        'builtin': _operators.abs,
        'forward': _operators.__abs__,
        'aliases': [
            'absolute',
        ],
    },
    'pos': {
        'builtin': _operators.pos,
        'forward': _operators.__pos__,
        'aliases': [
            'positive',
        ],
    },
    'neg': {
        'builtin': _operators.neg,
        'forward': _operators.__neg__,
        'aliases': [
            'negative',
        ],
    },
    'round': {
        'builtin': _operators.round,
        'forward': _operators.__round__,
    },
    'floor': {
        'builtin': _operators.floor,
        'forward': _operators.__floor__,
    },
    'ceil': {
        'builtin': _operators.ceil,
        'forward': _operators.__ceil__,
        'aliases': [
            'ceiling',
        ],
    },
    'trunc': {
        'builtin': _operators.trunc,
        'forward': _operators.__trunc__,
        'aliases': [
            'truncate',
        ],
    },
    'add': {
        'builtin': _operators.add,
        'forward': _operators.__add__,
        'reverse': _operators.__radd__,
        'inplace': _operators.__iadd__,
        'aliases': [
            'addition',
        ],
    },
    'sub': {
        'builtin': _operators.sub,
        'forward': _operators.__sub__,
        'reverse': _operators.__rsub__,
        'inplace': _operators.__isub__,
        'aliases': [
            'subtraction',
            'subtract',
        ],
    },
    'mul': {
        'builtin': _operators.mul,
        'forward': _operators.__mul__,
        'reverse': _operators.__rmul__,
        'inplace': _operators.__imul__,
        'aliases': [
            'multiplication',
            'multiply',
        ],
    },
    'truediv': {
        'builtin': _operators.truediv,
        'forward': _operators.__truediv__,
        'reverse': _operators.__rtruediv__,
        'inplace': _operators.__itruediv__,
        'aliases': [
            'true divide',
            'true_divide',
            'divide',
            'division',
            'true division',
            'floating-point division',
        ],
    },
    'floordiv': {
        'builtin': _operators.floordiv,
        'forward': _operators.__floordiv__,
        'reverse': _operators.__rfloordiv__,
        'inplace': _operators.__ifloordiv__,
        'aliases': [
            'floor divide',
            'floor_divide',
            'floor division',
        ],
    },
    'mod': {
        'builtin': _operators.mod,
        'forward': _operators.__mod__,
        'reverse': _operators.__rmod__,
        'inplace': _operators.__imod__,
        'aliases': [
            'mod',
            'remainder',
        ],
    },
    'pow': {
        'builtin': _operators.pow,
        'forward': _operators.__pow__,
        'reverse': _operators.__rpow__,
        'inplace': _operators.__ipow__,
        'aliases': [
            'power',
        ],
    },
}
_OPERATIONS = aliased.Mapping(_OPERATION_METADATA)
"""An aliased collection of operations."""


IMPLEMENTATIONS = aliased.Mapping(
    {
        (k, *v.get('aliases', [])): [m for n, m in v.items() if n != 'aliases']
        for k, v in _OPERATION_METADATA.items()
    }
)
"""An aliased mapping of callable operator implementations."""


FUNCTIONS = aliased.Mapping(
    {
        (k, *v.get('aliases', [])): v['builtin']
        for k, v in _OPERATION_METADATA.items()
    }
)
"""An aliased mapping of operator functions."""


METHODS = aliased.Mapping(
    {
        (k, *v.get('aliases', [])): [
            m for n, m in v.items() if n not in {'aliases', 'builtin'}
        ]
        for k, v in _OPERATION_METADATA.items()
    }
)
"""An aliased mapping of operator dunder methods."""


_DUNDER_MAP = {
    '__int__': _operators.__int__,
    '__float__': _operators.__float__,
    '__complex__': _operators.__complex__,
    '__eq__': _operators.__eq__,
    '__ne__': _operators.__ne__,
    '__lt__': _operators.__lt__,
    '__gt__': _operators.__gt__,
    '__le__': _operators.__le__,
    '__ge__': _operators.__ge__,
    '__abs__': _operators.__abs__,
    '__pos__': _operators.__pos__,
    '__neg__': _operators.__neg__,
    '__round__': _operators.__round__,
    '__floor__': _operators.__floor__,
    '__ceil__': _operators.__ceil__,
    '__trunc__': _operators.__trunc__,
    '__add__': _operators.__add__,
    '__radd__': _operators.__radd__,
    '__sub__': _operators.__sub__,
    '__rsub__': _operators.__rsub__,
    '__mul__': _operators.__mul__,
    '__rmul__': _operators.__rmul__,
    '__truediv__': _operators.__truediv__,
    '__rtruediv__': _operators.__rtruediv__,
    '__floordiv__': _operators.__floordiv__,
    '__rfloordiv__': _operators.__rfloordiv__,
    '__mod__': _operators.__mod__,
    '__rmod__': _operators.__rmod__,
    '__pow__': _operators.__pow__,
    '__rpow__': _operators.__rpow__,
}

DUNDER = container.Bijection(_DUNDER_MAP)


_CALLABLE_MAP = {k: v['builtin'] for k, v in _OPERATION_METADATA.items()}

CALLABLES = container.Bijection(_CALLABLE_MAP)


_OPERATOR_GROUPS = {
    'cast': ['int', 'float', 'complex'],
    'unary': ['abs', 'pos', 'neg'],
    'ordering': ['lt', 'le', 'gt', 'ge'],
    'additive': ['add', 'sub'],
    'multiplicative': ['mul', 'truediv'],
    'math': ['floor', 'ceil', 'trunc'],
}
_OPERATOR_GROUPS['comparative'] = [
    *_OPERATOR_GROUPS['ordering'],
    'eq',
    'ne',
]
_OPERATOR_GROUPS['algebraic'] = [
    *_OPERATOR_GROUPS['additive'],
    *_OPERATOR_GROUPS['multiplicative'],
    'pow',
]
_OPERATOR_GROUPS['binary'] = [
    *_OPERATOR_GROUPS['comparative'],
    *_OPERATOR_GROUPS['additive'],
    *_OPERATOR_GROUPS['multiplicative'],
]
_OPERATOR_GROUPS['real'] = [
    *_OPERATOR_GROUPS['comparative'],
    *_OPERATOR_GROUPS['algebraic'],
    *_OPERATOR_GROUPS['unary'],
    'floordiv',
    'mod',
]
_OPERATOR_GROUPS['scalar'] = [
    *_OPERATOR_GROUPS['real'],
    'round',
]


def _build_operators():
    """Internal function for building the default collection of operators.
    
    This function exists solely to reduce the scope of intermediate objects. It
    is not for public use.

    See Also
    --------
    `~OPERATORS`
        An immutable aliased mapping of default operators.
    `~get`
        A function for retrieving operators or operator groups by name.
    """
    _builtins = {
        key: operator['builtin']
        for key, operator in _OPERATIONS.items(aliased=True)
    }
    _operators = aliased.MutableMapping(_builtins)
    for name, keys in _OPERATOR_GROUPS.items():
        _operators.group(name, *keys)
    return _operators


class Mapping(aliased.Mapping):
    """A callable aliased mapping of numeric operators."""

    def __init__(self, mapping: typing.Mapping):
        if isinstance(mapping, aliased.MutableMapping):
            super().__init__(mapping.freeze(groups=True))
            self._groups = mapping._groups.copy()
        else:
            super().__init__(mapping)

    def callgroup(self, key: str, *args, **kwargs):
        """Compute the result of each operation in a group."""
        if key not in self._groups:
            raise KeyError(
                f"Unknown operator group: {key!r}"
            ) from None
        names = self._groups[key]
        operators = self[key]
        pairs = zip(names, operators)
        if len(args) == 1:
            for name in names:
                if name not in self._groups['unary']:
                    raise ValueError(
                        f"Cannot compute non-unary operation {name!r}"
                        " to a single argument"
                    ) from None
            arg = args[0]
            return {k: f(arg, **kwargs) for (k, f) in pairs}
        return {
            k: self._compute(k, f, *args, **kwargs)
            for (k, f) in zip(names, operators)
        }

    def _compute(self, name, f, *args, **kwargs):
        """Helper for `~callgroup`."""
        if name in self._groups['unary']:
            return tuple(f(arg, **kwargs) for arg in args)
        return f(*args, **kwargs)

    def __call__(self, key: str, *args, **kwargs):
        """Compute the result of an operation on the given arguments."""
        try:
            r = self[key]
        except KeyError as err:
            raise ValueError(
                f"Unknown operator or group: {key!r}"
            ) from err
        if isinstance(r, tuple):
            return tuple([f(*args, **kwargs) for f in r])
        return r(*args, **kwargs)

    def generate(self, *names: str, excluded: typing.Iterable[str]=None):
        """Retrieve standard operators or operator groups by name.
        
        Parameters
        ----------
        *names : strings
            The names of requested operators or operator groups. If missing,
            this function will produce all standard operators.

        excluded : iterable of strings, optional
            Names of operators to exclude.

        Notes
        -----
        The available operator groups and their corresponding operators are

        * `'unary'`: `['abs', 'pos', 'neg']`,
        * `'comparative'`: `['eq', 'ne', 'lt', 'le', 'gt', 'ge']`,
        * `'additive'`: `['add', 'sub']`,
        * `'multiplicative'`: `['mul', 'truediv', 'floordiv', 'mod']`,

        See Also
        --------
        `~OPERATORS`
            An immutable aliased mapping of default operators.
        """
        if not names:
            yield from self.values()
        keys = []
        for name in names:
            if name in self._groups:
                keys.extend(self._groups[name])
            else:
                keys.append(name)
        values = (self[key] for key in set(keys) - set(excluded or ()))
        yield from values


_OPERATORS = Mapping(_build_operators())
"""An aliased collection of canonical numeric operators."""


@typing.overload
def resolve(
    operation,
    strict: typing.Optional[bool],
) -> typing.Optional[typing.Callable]: ...

@typing.overload
def resolve(
    operation,
    strict: typing.Literal[True],
) -> typing.Callable: ...

def resolve(operation, strict: bool=False):
    """Get the callable operator corresponding to `operation`."""
    if not (isinstance(operation, str) or callable(operation)):
        raise TypeError(operation)
    if operation in _OPERATORS.values():
        return operation
    if found := _OPERATORS.get(operation):
        return found
    if strict:
        raise ValueError(
            f"Cannot determine appropriate operator for {operation!r}"
        ) from None


class OperandTypeError(TypeError):
    """Wrong operand type for a numeric operation."""





import argparse

from . import cli
from . import parameter


def print_version():
    from importlib.metadata import version
    print(f"{__package__} {version(__package__)}")


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        prog=__package__,
        description="Command-line interface for EPREM Python utilities",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        '-V', '--version',
        help="print the current package version",
        action='store_true',
    )
    subparsers = parser.add_subparsers(
        title="modes",
        dest="mode",
    )
    configfile_parser = subparsers.add_parser(
        'configfile',
        description=cli.DESCRIPTIONS['configfile'],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        help="utilities for EPREM runtime configuration files",
    )
    configfile_parser.add_argument(
        '-s', '--source',
        help=(
            "use source code in SRC for default parameter values"
            f"; otherwise, use the {__package__} database of default values"
        ),
        metavar='SRC',
    )
    configfile_mode = configfile_parser.add_mutually_exclusive_group(
        required=True,
    )
    configfile_mode.add_argument(
        '-g', '--generate',
        help="create a configuration file from default values",
        metavar='FILE',
    )
    configfile_mode.add_argument(
        '-c', '--compare',
        dest='files',
        help="compare values in the configuration file(s) to default values",
        nargs='+',
        metavar=('FILE0', 'FILE1'),
    )
    configfile_parser.add_argument(
        '-v', '--verbose',
        help="(generate) print status messages",
        action='store_true',
    )
    configfile_parser.add_argument(
        '--diff',
        help="(compare) show only parameters with differing values",
        action='store_true',
    )
    configfile_parser.add_argument(
        '--names',
        help="(compare) show full file names",
        action='store_true',
    )
    options = vars(parser.parse_args())
    if options.pop('version', None):
        print_version()
    elif usermode := options.pop('mode'):
        try:
            cli.run(usermode, options)
        except cli.UserError as err:
            print()
            print(f"ERROR: {err}")
            print(f"       Try 'python -m {__package__} {usermode} --help'")
    else:
        parser.error(f"No arguments. Try 'python -m {__package__} --help'")




class ParsingTypeError(TypeError):
    """An argument to `~parse` has an invalid type."""


class ParsingValueError(ValueError):
    """An argument to `~parse` has an invalid value."""


class MeasurementTypeError(TypeError):
    """An object's `__measure__` method did not return a `~Measurement`."""


class MeasuringTypeError(TypeError):
    """An argument to `~measure` has an invalid type."""





import pathlib
import typing

import matplotlib.pyplot as plt
import numpy
import numpy.typing

from .. import metric
from .. import measured
from .. import numeric
from .. import quantity
from .. import real
from ..typehelp import Self
from ._axis import Coordinates
from ._axes import Axes
from ._tensor import Tensor
from ._types import QuantityType as Q


T = typing.TypeVar('T')


class NDimError(Exception):
    """Error related to array dimension."""


class Array(Tensor[real.Array[real.ValueType]]):
    """A physical tensor with labeled axes."""

    def __init__(
        self,
        data: real.Array[real.ValueType],
        context: measured.Context[Q],
        axes: Axes,
        /,
    ) -> None:
        self._data_interface = data
        super().__init__(data, context)
        self._axes = axes
        self._scale = 1.0
        self._squeezed = None
        self._scaled = None

    def __str__(self) -> str:
        attrs = f"unit={str(self.unit)!r},\ndimensions={self.dimensions}"
        return f"{self.data.array},\n{attrs}"

    def __int__(self):
        """Called for int(self)."""
        return self._as_builtin(int)

    def __float__(self):
        """Called for float(self)."""
        return self._as_builtin(float)

    def __complex__(self):
        """Called for complex(self)."""
        return complex(float(self))

    def _as_builtin(self, t: T):
        """Convert this instance to a built-in numeric type."""
        if self.size == 1:
            return t(self.data.array.ravel()[0])
        raise TypeError(
            f"Cannot convert a size-{self.size} array to {t!r}"
        ) from None

    def __getitem__(self, args, /):
        idx = numeric.index.expand(self.ndim, args)
        data = self.data[idx]
        axes = {k: v[i] for i, (k, v) in zip(idx, self.axes.items())}
        return self.spawn(data, unit=self.unit, axes=axes)

    @typing.overload
    def transpose(
        self,
        axes: typing.Iterable[typing.Union[str, typing.SupportsIndex]],
        /,
    ) -> Self: ...

    @typing.overload
    def transpose(
        self,
        *axes: typing.Union[str, typing.SupportsIndex],
    ) -> Self: ...

    def transpose(self, *args):
        """Transpose this array."""
        if not args:
            return self
        data = self._data_interface.transpose(*args)
        axes = self.axes.permute(*args)
        return self.spawn(data, unit=self.unit, axes=axes)

    @typing.overload
    def plot(
        self,
        *args,
        x: typing.Optional[typing.Union[str, typing.Sequence]]=None,
        unit: typing.Optional[metric.UnitLike]=None,
        path: typing.Optional[pathlib.Path]=None,
        **extra
    ) -> None: ...

    def plot(self, *args, **kwargs):
        """Plot this array."""
        return plot(self, *args, **kwargs)

    @property
    def squeezed(self):
        """The equivalent `numpy` array, with singular dimensions removed."""
        if self._squeezed is None:
            self._squeezed = numpy.array(self).squeeze()
        return self._squeezed

    @property
    def data(self):
        if self._scaled is None:
            self._scaled = self._scale * self._data_interface
        return self._scaled

    @property
    def shapemap(self):
        """A mapping from dimension to axis length."""
        return self._data_interface.shapemap

    @property
    def dimensions(self):
        """This array's dimensions."""
        return self.axes.dimensions

    @property
    def axes(self) -> Axes:
        """This array's axes."""
        return self._axes

    _FUNCTIONS = {}
    _OPERATORS = {}


def converter(c: metric.Conversion, x: Array) -> Array:
    """Unit converter for physical arrays.

    Notes
    -----
    - This function differs from the unit-converter used by other `~physical`
      subclasses of `~measured.Object` by first creating an intermediate
      instance of `~Array` that contains a copy of the existing data and axes
      along with the new unit. It then updates the internal `_scale` attribute
      of the new instance so that calls to the `array` property will produce
      values consistent with the new unit.
    """
    if c.old == c.new:
        return x
    new = x.spawn(x._data_interface, unit=c.new, axes=x.axes)
    new._scale = float(c)
    return new


class AxisError(Exception):
    """Invalid plotting axis."""


_AXES_ARG_GROUPS = (
    'label',
    'lim',
    'scale',
    'bound',
    'margin',
    'ticklabels',
    'ticks',
)
_AXES_ARG_KEYS = tuple(
    [f'{q}{l}' for l in _AXES_ARG_GROUPS for q in {'x', 'y'}]
) + (
    'title',
)

def plot(
    array: Array,
    *args,
    x: typing.Optional[typing.Union[str, typing.Sequence]]=None,
    unit: typing.Optional[metric.UnitLike]=None,
    path: typing.Optional[pathlib.Path]=None,
    **extra
) -> None:
    """Plot the named array."""
    ndarray = numpy.array(array.withunit(unit or array.unit)).squeeze()
    if ndarray.ndim != 1:
        raise NDimError(f"The given array is not logically 1-D") from None
    extra['ylabel'] = f"[{array.unit.format(style='tex')}]"
    if x is None:
        dimension = next(i for i, n in array.shapemap.items() if n != 1)
        axis = array.axes[dimension]
        name = dimension.capitalize()
        if isinstance(axis, Coordinates):
            m = quantity.measure(axis)
            xarr = numpy.array(m)
            if 'xlabel' not in extra:
                extra['xlabel'] = f"{name} [{m.unit}]"
        else:
            xarr = numpy.array(axis)
            extra['xlabel'] = name
        _create_plot(xarr, ndarray, *args, **extra)
    else:
        if isinstance(x, measured.Type):
            xarr = numpy.array(x).squeeze()
        elif isinstance(x, Coordinates):
            m = quantity.measure(x)
            xarr = numpy.array(m)
            if 'xlabel' not in extra:
                extra['xlabel'] = f"{name} [{m.unit}]"
        elif isinstance(x, str):
            axis = array.axes[x]
            xarr = numpy.array(axis)
            extra['xlabel'] = x
        else:
            raise TypeError(x)
        _create_plot(xarr, ndarray, *args, **extra)
    if path is None:
        plt.show()
    else:
        plt.savefig(path)
    plt.close()


def _create_plot(*args, **extra):
    """Helper for `~plot`."""
    axes_kw = {}
    plot_kw = {}
    for key, value in extra.items():
        if key in _AXES_ARG_KEYS:
            axes_kw[key] = value
        else:
            plot_kw[key] = value
    plt.subplot(**axes_kw)
    plt.plot(*args, **plot_kw)


